# SCO UNIX V/386 (release 3.2) makefile for PC-KIMMO
#					Steve McConnel,  14-Jul-90
#		add shar pseudo-target, Steve McConnel,   1-Feb-91
############################################################################
#
#		compilation rules
#
.SUFFIXES: .om .og .obj

# use the AT&T compiler by default
CFLAGS= -O -DUNIX -DUSG
LINTFLAGS = -DUNIX -DUSG
CC=rcc

# use the GNU compiler on request
GU_CFLAGS= -DUNIX -DUSG -g -O -W -Wall -Wpointer-arith -Wcast-qual -Wshadow
.c.og:
	gcc $(GU_CFLAGS) -o $@ -c $<

# use the Microsoft compiler on request
MU_CFLAGS= -DUNIX -DUSG -Octl
.c.om:
	cc $(MU_CFLAGS) -Fo$@ -c $<

# compile for MS-DOS on request
MSDOS_CFLAGS= -dos -DMSC -M0le -Ot
.c.obj:
	cc $(MSDOS_CFLAGS) -Fo$@ -c $<

############################################################################

OBJS=pckimmo.o usercmd.o userfunc.o\
 lexicon.o rules.o generate.o recogniz.o pckfuncs.o

GU_OBJS=pckimmo.og usercmd.og userfunc.og\
 lexicon.og rules.og generate.og recogniz.og pckfuncs.og

MU_OBJS=pckimmo.om usercmd.om userfunc.om\
 lexicon.om rules.om generate.om recogniz.om pckfuncs.om

MSDOS_OBJS=pckimmo.obj usercmd.obj userfunc.obj\
 lexicon.obj rules.obj generate.obj recogniz.obj pckfuncs.obj

SOURCES=pckimmo.c usercmd.c userfunc.c\
 lexicon.c rules.c generate.c recogniz.c pckfuncs.c
HEADERS=pckimmo.h version.h
EXTRAFILES=Errata Makefile Makefile.MSC Makefile.SCO README g.c r.c

pckimmo: $(OBJS)
	$(CC) $(CFLAGS) -o pckimmo $(OBJS)

pckimmo-g: $(GU_OBJS)
	gcc $(GU_CFLAGS) -o pckimmo-g $(GU_OBJS)

pckimmo-m: $(MU_OBJS)
	cc $(MU_CFLAGS) -o pckimmo-m $(MU_OBJS)

pckimmo4.exe: $(MSDOS_OBJS)
	cc $(MSDOS_CFLAGS) -F 4000 -o pckimmo4.exe $(MSDOS_OBJS)

pckimmo8.exe: $(MSDOS_OBJS)
	cc $(MSDOS_CFLAGS) -F 8000 -o pckimmo8.exe $(MSDOS_OBJS)

all: pckimmo pckimmo-g pckimmo-m pckimmo4.exe pckimmo8.exe g g.exe r r.exe
	@echo pckimmo, pckimmo-g, pckimmo-m, and 'pckimmo[48].exe' have all been made

pckimmo.lint: $(SOURCES)
	lint $(LINTFLAGS) $(SOURCES) >pckimmo.lint

pckimmo.calls: $(SOURCES)
	calls $(CALLFLAGS) $(SOURCES) >pckimmo.calls

pckimmo.o  pckimmo.om  pckimmo.og  pckimmo.obj:  pckimmo.c  pckimmo.h version.h
usercmd.o  usercmd.om  usercmd.og  usercmd.obj:  usercmd.c  pckimmo.h
userfunc.o userfunc.om userfunc.og userfunc.obj: userfunc.c pckimmo.h

lexicon.o  lexicon.om  lexicon.og  lexicon.obj:  lexicon.c  pckimmo.h
rules.o    rules.om    rules.og    rules.obj:    rules.c    pckimmo.h
recogniz.o recogniz.om recogniz.og recogniz.obj: recogniz.c pckimmo.h
generate.o generate.om generate.og generate.obj: generate.c pckimmo.h
pckfuncs.o pckfuncs.om pckfuncs.og pckfuncs.obj: pckfuncs.c pckimmo.h

#
#  simple test programs to check out modularity
#
R_OBJS = r.o rules.o lexicon.o recogniz.o pckfuncs.o
R_OBJS_MS = r.obj rules.obj lexicon.obj recogniz.obj pckfuncs.obj

r: $(R_OBJS)
	$(CC) $(CFLAGS) -o r $(R_OBJS)
r.exe: $(R_OBJS_MS)
	cc $(MSDOS_CFLAGS) -F 4000 -o r.exe $(R_OBJS_MS)
r.o r.obj: r.c pckimmo.h

G_OBJS = g.o rules.o generate.o pckfuncs.o
G_OBJS_MS = g.obj rules.obj generate.obj pckfuncs.obj

g: $(G_OBJS)
	$(CC) $(CFLAGS) -o g $(G_OBJS)
g.exe: $(G_OBJS_MS)
	cc $(MSDOS_CFLAGS) -F 4000 -o g.exe $(G_OBJS_MS)
g.o g.obj: g.c pckimmo.h

shar: $(SOURCES) $(HEADERS) $(EXTRAFILES)
	split -1750 usercmd.c usercmd.c-
	shar -a -pX README Makefile generate.c lexicon.c >pckimmo.sh01
	shar -a -pX pckfuncs.c pckimmo.c pckimmo.h >pckimmo.sh02
	shar -a -pX Errata recogniz.c g.c r.c version.h usercmd.c-ab | \
	    sed '/^exit 0$$/d' >pckimmo.sh03
	echo 'if test -f usercmd.c-aa -a -f usercmd.c-ab' >>pckimmo.sh03
	echo '    then' >>pckimmo.sh03
	echo '    cat usercmd.c-aa usercmd.c-ab >usercmd.c' >>pckimmo.sh03
	echo '    echo shar: combining usercmd.c-aa and usercmd.c-ab into usercmd.c' >>pckimmo.sh03
	echo '    rm usercmd.c-aa usercmd.c-ab' >>pckimmo.sh03
	echo '    fi' >>pckimmo.sh03
	echo 'exit 0' >>pckimmo.sh03
	shar -a -pX Makefile.MSC Makefile.SCO rules.c >pckimmo.sh04
	shar -a -pX usercmd.c-aa | sed '/^exit 0$$/d' >pckimmo.sh05
	echo 'if test -f usercmd.c-aa -a -f usercmd.c-ab' >>pckimmo.sh05
	echo '    then' >>pckimmo.sh05
	echo '    cat usercmd.c-aa usercmd.c-ab >usercmd.c' >>pckimmo.sh05
	echo '    echo shar: combining usercmd.c-aa and usercmd.c-ab into usercmd.c' >>pckimmo.sh05
	echo '    rm usercmd.c-aa usercmd.c-ab' >>pckimmo.sh05
	echo '    fi' >>pckimmo.sh05
	echo 'exit 0' >>pckimmo.sh05
	shar -a -pX userfunc.c >pckimmo.sh06
	rm usercmd.c-aa usercmd.c-ab
	wc -c pckimmo.sh*

clean:
	rm -f *.o *.om *.og *.obj
	rm -f pckimmo pckimmo-m pckimmo-g pckimmo4.exe pckimmo8.exe r g
	rm -f pckimmo.lint pckimmo.calls
