/*	PCKIMMO.C - global variables and main() for PC-KIMMO program
 ***************************************************************************
 *
 *	int main(argc, argv)
 *	int argc;
 *	char **argv;
 *
 ***************************************************************************
 *	EDIT HISTORY
 *	18-Jul-89	written by Dave Smith
 *	18-Sep-89	Steve McConnel
 *			     - regularize some comments and includes
 *			     - clean up code in printResults() to free memory
 *			     - revised cmdLine() to use getopt()
 *	21-Sep-89	SRMc - added \v\f\r to whiteSpc[]
 *			     - change char to unsigned char for 8-bit safety
 *	23-Sep-89	SRMc - rename MORPH.H to KIMMO.H, MORPH.C to KIMMO.C
 *	25-Sep-89	SRMc - move printResults() to sh_util.c
 *			     - finetune screen displays
 *	26-Sep-89	SRMc - eliminate noneRslt structure
 *	28-Sep-89	SRMc - make Lang global, eliminate "lang" arguments
 *			     - simplify tracing and logging variables
 *	 2-Oct-89	SRMc - rename Language to LANGUAGE
 *			     - rename ResNode to RESULT
 *	 3-Oct-89	SRMc - rename file to PC-KIMMO.C
 *			     - reorganize source files for new user interface
 *				borrowed from my RT-11 Kermit program
 *			     - rename cmdLine() to parse_cmd_line()
 *	11-Oct-89	SRMc - add Timing_flag and SET TIMING command
 *	13-Oct-89	SRMc - add SET LIMIT {ON|OFF} command
 *	14-Oct-89	SRMc - imported getopt() from AT&T Public Domain
 *				GETOPT.C
 *			     - eliminate resTail as a global variable
 *	21-Oct-89	SRMc - ifdef getopt() with NOGETOPT rather than USG
 *	24-Oct-89	SRMc - some delinting
 *	 2-Jan-90	SRMc - adjust type definitions of global variables
 *			     - more delinting, add function prototypes
 *	 3-Jan-90	SRMc - will we never run out of lint?
 *	24-Jan-90	SRMc - edit trace and error messages
 *	25-Jan-90	SRMc - add check for duplicated subset names
 *			     - edit the help messages
 *	26-Jan-90	SRMc - add automatic ".lex" extension for INCLUDE files
 *			     - edit trace messages
 *			     - port to 4.2BSD (SunOS)
 *	 8-Mar-90	EA   - #ifdef for THINK_C to use console library
 *	19-Apr-90	JT   - fixed comparison of strcmp with NULL to 0 in
 *				getopt
 *	12-Jul-90	SRMc - replace "void *" with "VOIDP", as suggested
 *				by Greg Lee (lee@uhccux.uhcc.hawaii.edu) for
 *				port to ULTRIX
 *	14-Jul-90	SRMc - Version 1.0A banner
 *	 6-Sep-90	SRMc - Version 1.0B banner (fixed bugs reported by
 *				Greg Lee and Richard Sproat)
 *	12-Sep-90	SRMc - Version 1.0.3 banner (fixed bug reported by
 *				by Richard Sproat related to one reported by
 *				Greg Lee that I thought I'd fixed...)
 *			     - go to using version.h for version numbers and
 *				dates.  fix so that banner won't have to be
 *				changed for every bug fix.
 *	26-Feb-91	SRMc - add 1991 to copyright dates
 ***************************************************************************
 * Copyright 1989, 1990, 1991 by the Summer Institute of Linguistics, Inc.
 * All rights reserved.
 */
#include <stdio.h>
#include <ctype.h>
#include <setjmp.h>
#include "pckimmo.h"
#include "userfunc.h"
#include "version.h"

#ifdef THINK_C
#include <console.h>
#include <unix.h>
#endif

/* standard library functions */
#ifdef __STDC__
extern void exit(int status);
#else
extern void exit();
#endif

/***************************************************************************
 * GLOBAL VARIABLES
 */
LANGUAGE Lang;			/* data for the current language */
int Limit_flag = 0;		/* flag for limit on/off */
int Timing_flag = 0;		/* flag for timing on/off */
int Trace_flag = 0;		/* flag for tracing on/off */
int Verbose_flag = 0;		/* flag for verbose comparison on/off */
FILE *Log_fp = (FILE *)NULL;	/* log FILE pointer */
char *log_file = (char *)NULL;	/* current log file */
jmp_buf menu_env;		/* environment to restore for error recovery */
				/* same chars as isspace() */
unsigned char whiteSpc[] = { ' ', '\t', '\n', '\v', '\f', '\r', '\0' };
/*
 *  VARIABLES SET BY COMMAND LINE OPTIONS
 */
unsigned char commentChr = ';';
char *rule_file = (char *)NULL;		/* current rule file */
char *lex_file  = (char *)NULL;		/* current lexicon file */
char *cmd_file  = (char *)NULL;		/* file of commands */

#ifdef NOGETOPT
#ifdef BSD
#include <strings.h>
#else
#include <string.h>
#endif
/*************************************************************************
 * Here's something you've all been waiting for:  the AT&T public domain
 * source for getopt(3).  It is the code which was given out at the 1985
 * UNIFORUM conference in Dallas.  I obtained it by electronic mail
 * directly from AT&T.  The people there assure me that it is indeed
 * in the public domain.
 *
 * There is no manual page.  That is because the one they gave out at
 * UNIFORUM was slightly different from the current System V Release 2
 * manual page.  The difference apparently involved a note about the
 * famous rules 5 and 6, recommending using white space between an option
 * and its first argument, and not grouping options that have arguments.
 * Getopt itself is currently lenient about both of these things White
 * space is allowed, but not mandatory, and the last option in a group can
 * have an argument.  That particular version of the man page evidently
 * has no official existence, and my source at AT&T did not send a copy.
 * The current SVR2 man page reflects the actual behavor of this getopt.
 * However, I am not about to post a copy of anything licensed by AT&T.
 */
static int	opterr = 1;
static int	optind = 1;
static int	optopt;
static char	*optarg;

static ERR(argv0, s, c)
char *argv0;
char *s;
char c;
{
static char errbuf[2];
(void) write(2, argv0, (unsigned)strlen(argv0));
(void) write(2, s, (unsigned)strlen(s));
errbuf[0] = c;
errbuf[1] = '\n';
(void) write(2, errbuf, 2);
}

static int getopt(argc, argv, opts)
int	argc;
char	**argv, *opts;
{
static int sp = 1;
register int c;
register char *cp;

if (sp == 1)
    {
    if (optind >= argc || argv[optind][0] != '-' || argv[optind][1] == '\0')
	return(EOF);
    else if(strcmp(argv[optind], "--") == 0)
	{
	optind++;
	return(EOF);
	}
    }
optopt = c = argv[optind][sp];
if (c == ':' || (cp=strchr(opts, c)) == NULL)
    {
    if (opterr)
	ERR(argv[0], ": illegal option -- ", c);
    if (argv[optind][++sp] == '\0')
	{
	optind++;
	sp = 1;
	}
    return('?');
    }
if (*++cp == ':')
    {
    if (argv[optind][sp+1] != '\0')
			optarg = &argv[optind++][sp+1];
    else if(++optind >= argc)
	{
	if (opterr)
	    ERR(argv[0], ": option requires an argument -- ", c);
	sp = 1;
	return('?');
	}
    else
	optarg = argv[optind++];
    sp = 1;
    }
else
    {
    if (argv[optind][++sp] == '\0')
	{
	sp = 1;
	optind++;
	}
    optarg = NULL;
    }
return(c);
}

#else /*NOGETOPT*/

#ifdef __STDC__
extern int getopt(int argc, char **argv, char *optstring);
#else
extern int getopt();
#endif

extern int optind;
extern char *optarg;


#endif /*NOGETOPT*/

/****************************************************************************
 * NAME
 *    parse_cmd_line
 * ARGUMENTS
 *    argc - number of command line arguments
 *    argv - pointer to array of command line arguments
 * DESCRIPTION
 *    Handle the command line arguments for PC-KIMMO.
 * RETURN VALUE
 *    none
 */
static void parse_cmd_line( argc, argv )
int argc;
char **argv;
{
int k;
int errflag = 0;

while ((k = getopt(argc, argv, "C:c:L:l:R:r:T:t:")) != EOF)
    {
    switch (k)
	{
	case 'C':		/* comment character */
	case 'c':
	    commentChr = *optarg;
	    break;

	case 'L':		/* initial lexicon filename */
	case 'l':
	    lex_file = optarg;
	    break;

	case 'R':		/* initial rules filename */
	case 'r':
	    rule_file = optarg;
	    break;

	case 'T':		/* initial command filename */
	case 't':
	    cmd_file = optarg;
	    break;

	default:
	    ++errflag;
	    break;
	}
    }
if (errflag)
    {
    fputs("\
Usage:  pc-kimmo [-c char] [-r rulefile] [-l lexfile] [-t cmdfile]\n\
    -c char      make char the comment character (default is ;)\n\
    -r rulefile  read rules from rulefile\n\
    -l lexfile   read a lexicon from lexfile (requires -r also)\n\
    -t cmdfile   take commands from cmdfile\n\
", stderr );
    exit(2);
    }
}

/****************************************************************************
 * NAME
 *    main
 * ARGUMENTS
 *    argc - number of command line arguments
 *    argv - pointer to array of command line arguments
 * DESCRIPTION
 *    main() for PC-KIMMO -- where the program starts executing
 * RETURN VALUE
 *    status for return: 0 if okay, nonzero if error
 */
int main( argc, argv )
int argc;
char **argv;
{
fprintf(stderr,"PC-KIMMO TWO-LEVEL PROCESSOR\n");
fprintf(stderr,"Version %d.%d.%d (%s), Copyright %s SIL\n",
				VERSION, REVISION, PATCHLEVEL, DATE, YEAR );
#ifdef DJGPP
fprintf(stderr,"Beta version for 386 CPU under MS-DOS (VCPI compliant)\n");
#endif

#ifdef THINK_C
SetWTitle( FrontWindow(), "\pPC-KIMMO");
argc = ccommand(&argv);
#endif

parse_cmd_line(argc,argv);  /* parse the command line */
user_cmd();			/* interact with the user */
return(0);
}
