/*	USERCMD.C - interactive user interface for PC-KIMMO
 ***************************************************************************
 *
 *	void user_cmd()
 *
 ***************************************************************************
 *	EDIT HISTORY
 * (RT-11 Kermit user interface based on work by Frank da Cruz, CUCCA)
 *	 1-Jun-85	SRMc - initial edit from Unix CKermit sources
 *	   ...			...
 *	11-Dec-87	SRMc - final revision of RT-11/TSX Kermit
 *	   ...			...
 *	 3-Oct-89	SRMc - copy code, revise it for use in PC-KIMMO
 *				program (total rewrite except for struct
 *				definition and bottom-level parsing funcs)
 *	 4-Oct-89	SRMc - clean up some rough edges of the new interface
 *			     - rename buildLex() to load_lexicon(), check for
 *				error returns
 *	 5-Oct-89	SRMc - revise file name parsing to use set_filename()
 *	10-Oct-89	SRMc - use report_error() for error messages
 *	11-Oct-89	SRMc - revise storage of subsets to use SUBSET struct
 *			     - add default name and type for TAKE files
 *			     - add SET TIMING command
 *	12-Oct-89	SRMc - handle default extension for -t takefile
 *				command line option
 *			     - extract code to userfunc.c to write
 *				compare_recog() and compare_gener()
 *	13-Oct-89	SRMc - if "-l lexfile" without "-r rulefile", erase
 *				the invalid lex_file value
 *			     - add SET LIMIT {ON|OFF} command
 *	16-Oct-89	SRMc - edit help and error messages
 *	17-Oct-89	SRMc - call free_lexicons() before load_lexicon()
 *	20-Oct-89	SRMc - rename getRules() to load_rules(), restore
 *				(LANGUAGE *) argument
 *			     - simplified error messages
 *	21-Oct-89	SRMc - new interface to generator() and recognizer()
 *			     - rename load_lexicon() to load_lexicons()
 *			     - new interface to getline(), load_rules(), and
 *				load_lexicons()
 *	24-Oct-89	SRMc - some delinting
 *	13-Dec-89	SRMc - add filename to report_error() argument list
 *			     - change .ls extension to .pai for FILE COMPARE
 *				PAIRS command
 *			     - display dots on screen for FILE GENERATE or
 *				FILE RECOGNIZE if output going to a file
 *			     - add the EDIT command
 *			     - edit the help messages
 *	15-Dec-89	SRMc - fix bug in dots display of FILE GENERATE and
 *				FILE RECOGNIZE
 *	 2-Jan-90	SRMc - more delinting, add function prototypes
 *	 3-Jan-90	SRMc - will we never run out of lint?
 *	25-Jan-90	SRMc - edit the help messages
 *	 8-Mar-90	EA   - conditional compile do_edit, do_system, timing
 *				for THINK C on Macintosh
 *	 7-May-90	JT   - added CD command for Mac version
 *	12-Jul-90	SRMc - replace "void *" with "VOIDP", as suggested
 *				by Greg Lee (lee@uhccux.uhcc.hawaii.edu) for
 *				port to ULTRIX
 *	 6-Sep-90	SRMc - #ifdef HELP message for Macintosh (THINK C)
 *	19-Nov-91	SRMc - add fflush(stderr) for THINK C following each
 *				putc('.',stderr)
 *	 5-Dec-91	SRMc - add prototypes for Think C 5.0
 *	18-Feb-92	SRMc - add typecasts to satisfy THINK C
 ***************************************************************************
 * Copyright 1989, 1992 by the Summer Institute of Linguistics, Inc.
 * All rights reserved.
 */
#include <stdio.h>
#include <ctype.h>
#include <setjmp.h>
#include "pckimmo.h"
#include "userfunc.h"

#ifdef BSD
#include <strings.h>
#else
#include <string.h>
#endif

#ifdef USE_SIGNAL
#include <signal.h>
#ifdef __STDC__
extern void sigint_handler(int sig);
#else
extern int sigint_handler();
#endif
#endif

#ifdef __STDC__
/* standard library functions */
#ifndef DJGPP
extern int atoi(char *str);
#endif


#else

/* standard library functions */
extern int atoi();

#endif

extern LANGUAGE Lang;		/* global data for the current language */
extern unsigned char commentChr; /* global comment character */
extern int Limit_flag;		/* global flag for limit on/off */
extern int Trace_flag;		/* global flag for tracing on/off */
extern int Verbose_flag;	/* global flag for verbose compare on/off */
extern int Timing_flag;		/* global flag for timing on/off */
extern FILE *Log_fp;		/* global log FILE pointer */
extern jmp_buf menu_env;    /* environment to restore for error recovery */

extern char *rule_file, *lex_file, *cmd_file, *log_file;
extern unsigned char whiteSpc[];

FILE *Fileout_fp;		/* needed by sigint_handler() */
/*
 *  error messages
 */
static struct message Ambig_command =
    { 101, "Ambiguous command: %s" };
static struct message Bad_command =
    { 102, "Invalid command: %s" };
static struct message Miss_keyword =
    { 103, "Missing keyword in %s command" };
struct message Miss_argument =		/* used in userfunc.c */
    { 104, "Missing argument in %s command" };
static struct message Ambig_keyword =
    { 105, "Ambiguous keyword in %s command: %s" };
static struct message Bad_keyword =
    { 106, "Invalid keyword in %s command: %s" };
struct message Bad_argument =		/* used in userfunc.c */
    { 107, "Invalid argument in %s command: %s" };
static struct message Miss_infile =
    { 108, "Missing input file argument in %s command" };
static struct message Bad_infile =
    { 109, "Cannot open input file %s in %s command" };
struct message Bad_outfile =		/* used in userfunc.c */
    { 110, "Cannot open output file %s in %s command" };
static struct message Need_rules_first =
    { 111, "Must load rules before loading lexicon" };
static struct message Too_deep_TAKE =
    { 112, "TAKE files nested too deeply" };
static struct message Error_in_TAKE =
    { 113, "TAKE file aborted due to invalid command: %s" };

/*************************************************************************
 *  Symbols local to this module
 */
#define CM_INV 1			/* Invisible keyword */
/*
 *  Values for invalid keyword searches
 */
#define LLNULL -4			/* Null keyword */
#define LLAMBI -3			/* Ambiguous keyword */
#define LLNOT  -2			/* Not a keyword */
#define LLHELP -1			/* '?' */
/*
 *  Values for successful keyword searches (must be >= 0)
 */
#define KW_ALL		 1
#define KW_CLOSE	 2
#define KW_COMMENT	 3
#define KW_COMPARE	 4
#define KW_EDIT		 5
#define KW_FILE		 6
#define KW_GENERATE	 7
#define KW_HELP		 8
#define KW_LEXICON	 9
#define KW_LIMIT	10
#define KW_LIST		11
#define KW_LOAD		12
#define KW_LOG		13
#define KW_NEW		14
#define KW_OFF		15
#define KW_ON		16
#define KW_PAIRS	17
#define KW_QUIT		18
#define KW_RECOGNIZE	19
#define KW_RULES	20
#define KW_SET		21
#define KW_SHOW		22
#define KW_STATUS	23
#define KW_SYSTEM	24
#define KW_TAKE		25
#define KW_TIMING	26
#define KW_TRACING	27
#define KW_VERBOSE	28
#ifdef THINK_C
#define KW_CD		29
#endif

/*************************************************************************
 * NAME
 *    struct keytab
 * DESCRIPTION
 *    Keyword Table Template
 */
struct keytab
    {
    char *kwd;		/* Pointer to keyword string */
    int val;		/* Associated value */
    int flgs;		/* Flags (as defined above) */
    };

/*************************************************************************
 *  Variables and symbols local to this module
 */
static char *prompt = "PC-KIMMO>";

#define MAXTAKE 3		/* Maximum nesting of TAKE files */
static FILE *tfile[MAXTAKE];	/* File pointers for TAKE command */
static int tlinenum[MAXTAKE];	/* line numbers for TAKE command files */
static int tlevel = -1;		/* Take command level */

/*************************************************************************
 *  keyword tables for interactive commands
 */
static struct keytab cmd_tab[] = {	/* Top-Level Keyword Table */
#ifdef THINK_C
	"cd",		KW_CD, 0,
#endif
    "close",		KW_CLOSE, 0,
    "compare",		KW_COMPARE, 0,
#ifndef THINK_C
    "edit",		KW_EDIT, 0,
#endif
    "exit",		KW_QUIT, 0,
    "file",		KW_FILE, 0,
    "generate",		KW_GENERATE, 0,
    "help",		KW_HELP, 0,
    "l",		KW_LOAD, CM_INV,
    "list",		KW_LIST, 0,
    "load",		KW_LOAD, 0,
    "log",		KW_LOG, 0,
    "new",		KW_NEW, 0,
    "quit",		KW_QUIT, 0,
    "recognize",	KW_RECOGNIZE, 0,
    "set",		KW_SET, 0,
    "show",		KW_SHOW, 0,
    "status",		KW_STATUS, 0,
#ifndef THINK_C
    "system",		KW_SYSTEM, 0,
#endif
    "take",		KW_TAKE, 0,
    };
static int n_cmd = (sizeof(cmd_tab) / sizeof(struct keytab));

static struct keytab load_tab[] = {	/* LOAD command keyword table */
    "lexicon",		KW_LEXICON, 0,
    "rules",		KW_RULES, 0,
    };
static int n_load = (sizeof(load_tab) / sizeof(struct keytab));

static struct keytab file_tab[] = {	/* FILE command keyword table */
    "compare",		KW_COMPARE, 0,
    "generate",		KW_GENERATE, 0,
    "recognize",	KW_RECOGNIZE, 0,
    };
static int n_file = (sizeof(file_tab) / sizeof(struct keytab));

static struct keytab set_tab[] = {	/* SET command keyword table */
    "comment",		KW_COMMENT, 0,
    "limit",		KW_LIMIT, 0,
    "rules",		KW_RULES, 0,
    "timing",		KW_TIMING, 0,
    "tracing",		KW_TRACING, 0,
    "verbose",		KW_VERBOSE, 0,
};
static int n_set = (sizeof(set_tab) / sizeof(struct keytab));

static struct keytab list_tab[] = {	/* LIST command keyword table */
    "lexicon",		KW_LEXICON, 0,
    "pairs",		KW_PAIRS, 0,
    "rules",		KW_RULES, 0,
    };
static int n_list = (sizeof(list_tab) / sizeof(struct keytab));

static struct keytab show_tab[] = {	/* SHOW command keyword table */
    "lexicon",		KW_LEXICON, 0,
    "rule",		KW_RULES, 0,
    "status",		KW_STATUS, 0,
    };
static int n_show = (sizeof(show_tab) / sizeof(struct keytab));

static struct keytab compare_tab[] = {	/* COMPARE command keyword table */
    "generate",		KW_GENERATE, 0,
    "pairs",		KW_PAIRS, 0,
    "recognize",	KW_RECOGNIZE, 0,
    };
static int n_compare = (sizeof(compare_tab) / sizeof(struct keytab));

static struct keytab onoff_tab[] = {	/* ON and OFF keywords table */
    "off",		KW_OFF, 0,
    "on",		KW_ON, 0,
    };
static int n_onoff = 2;

#define HLPLW  78			/* Width of ?-help line */
#define HLPCW  19			/* Width of ?-help column */

static char hlpbuf[MAXLINELEN];		/* Help string buffer */
static int hw = HLPLW,		/* Help line width */
	   hc = HLPCW,		/* Help line column width */
	   hh,			/* Current help column number */
	   hx;			/* Current help line position */

/*****************************************************************************
 * NAME
 *    clrhlp
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Initialize/Clear the help line buffer
 * RETURN VALUE
 *    none
 */
static void clrhlp()
{
hlpbuf[0] = NUL;
hh = hx = 0;
}

/*****************************************************************************
 * NAME
 *    dmphlp
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Dump the help line buffer.
 * RETURN VALUE
 *    none
 */
static void dmphlp()
{
hlpbuf[hx++] = NUL;
fprintf(stderr, " %s\n",hlpbuf);
clrhlp();
}

/*****************************************************************************
 * NAME
 *    addhlp
 * ARGUMENTS
 *    s - pointer to a character string
 * DESCRIPTION
 *    Add a string to the help line buffer
 * RETURN VALUE
 *    none
 */
static void addhlp(s)
register char *s;
{
register int j;

hh++;					/* Count this column */
for (j = 0; j < hc; j++)
    {					/* Fill the column */
    if (*s != NUL)			/* First with chars from the string */
	hlpbuf[hx++] = *s++;
    else
	{
	if (hh < (hw / hc))		/* Then with spaces */
	    hlpbuf[hx++] = ' ';
	else
	    {
	    hlpbuf[hx++] = NUL;		/* If last column, no spaces. */
	    dmphlp();			/* Print it. */
	    return;
	    }
    	} 
    }
if (*s != NUL)			/* Still some chars left in string? */
    hlpbuf[hx-1] = '+';		/* Mark as too long for column. */
}

/*****************************************************************************
 * NAME
 *    lower
 * ARGUMENTS
 *    s - pointer to a character string
 * DESCRIPTION
 *    Convert a string to lowercase.
 * RETURN VALUE
 *    length of the converted string
 */
static int lower(s)
register char *s;
{
register int length;
length = 0;
while (*s)
    {
    if (isupper(*s))
	*s = tolower(*s);
    ++s;
    ++length;
    }
return(length);
}

/*****************************************************************************
 * NAME
 *    lookup
 * ARGUMENTS
 *    word  - the target string to look up in the table.
 *    table - a 'struct keytab' table.
 *    tsize - the number of elements in the table.
 *    xhlp  - preface string for help message
 *	    (if empty, then '?' is valid; if NULL, then no help at all)
 * DESCRIPTION
 *    Search the given table of keywords for the word.  The keyword table
 *    must be arranged in ascending alphabetical order, and all letters must
 *    be lowercase.
 *
 *    A match is successful if the target matches a keyword exactly, or if
 *    the target is a prefix of exactly one keyword.  It is ambiguous if the
 *    target matches two or more keywords from the table.
 *
 * RETURN VALUE
 *    the keyword's associated value (zero or greater) if found, or
 *      LLNULL  if nothing to look up (target was null),
 *      LLAMBI  if ambiguous,
 *      LLNOT   if not found.
 *      LLHELP  if '?' is found in word (after printing list of keywords)
 */
static int lookup(cmd,table,tsize,xhlp)
register char *cmd;
struct keytab table[];
int tsize;
char *xhlp;
{
register int i, v;
int cmdlen;
/*
 *  Lowercase & get length of target, if it's null return code LLNULL.
 */
if (cmd == (char *)NULL)
    return(LLNULL);
if ((((cmdlen = lower(cmd))) == 0) || (tsize < 1))
    return(LLNULL);
/*
 *  If '?' found in target, return code LLHELP
 */
if (strchr(cmd,'?'))
    {
    if (xhlp == (char *)NULL)
	return(LLHELP);
    if (*xhlp == NUL)
	fprintf(stderr, "One of the following:\n");
    else
	fprintf(stderr, "%s, one of the following:\n",xhlp);
    clrhlp();
    for (i = 0; i < tsize; i++)
	{
	if (!(table[i].flgs & CM_INV))
	    addhlp(table[i].kwd);
	}
    dmphlp();
    return(LLHELP);
    }
/*
 *  Not null, look it up
 */
for (i = 0; i < tsize-1; i++)
    {
    if	   (!strcmp(table[i].kwd,cmd) ||
           ((v = !strncmp(table[i].kwd,cmd,cmdlen)) &&
		strncmp(table[i+1].kwd,cmd,cmdlen)) )
	{
	return(table[i].val);
	}
    if (v)
	return(LLAMBI);
    }	
/*
 *  Last (or only) element
 */
if (!strncmp(table[tsize-1].kwd,cmd,cmdlen))
    return(table[tsize-1].val);
else
    return(LLNOT);
}

/*****************************************************************************
 * NAME
 *    do_compare
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Finish parsing and execute a FILE COMPARE (or COMPARE) command.
 * RETURN VALUE
 *    none
 */
static void do_compare()
{
register char *arg;
char *infile;
FILE *infp;

arg = (char *)strtok8(NULL,whiteSpc);
switch (lookup( arg, compare_tab, n_compare, ""))
    {
    case LLNULL:			/* empty command */
	report_error(FATAL, &Miss_keyword, (int *)NULL, (char *)NULL,
							"[FILE] COMPARE" );
	break;

    case LLAMBI:			/* ambiguous command */
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"[FILE] COMPARE",arg);
	break;

    case LLHELP:			/* ? */
	fprintf(stderr, "Type HELP [FILE] COMPARE for more help\n");
	break;

    case KW_GENERATE:		/* FILE COMPARE GENERATE <infile> */
	infile = (char *)strtok8(NULL,whiteSpc);
	if (wanthelp(infile,"COMPARE GENERATE [<file>] (default name is data.gen, default type is .gen)"))
	    break;
	infile = set_filename( infile, "data.gen", ".gen");
	infp = fopen( infile, "r");
	if (infp == (FILE *)NULL)
	    {
	    report_error(FATAL, &Bad_infile, (int *)NULL, (char *)NULL,
					infile, "[FILE] COMPARE GENERATE" );
	    myfree(infile);
	    return;
	    }
	compare_gener(infp);
	fclose(infp);
	myfree(infile);
	if (Timing_flag)
	    show_timing();
	break;

    case KW_RECOGNIZE:		/* FILE COMPARE RECOGNIZE <infile> */
	infile = (char *)strtok8(NULL,whiteSpc);
	if (wanthelp(infile,"COMPARE RECOGNIZE [<file>] (default name is data.rec, default type is .rec)"))
	    break;
	infile = set_filename( infile, "data.rec", ".rec");
	infp = fopen( infile, "r");
	if (infp == (FILE *)NULL)
	    {
	    report_error(FATAL, &Bad_infile, (int *)NULL, (char *)NULL,
				infile, "[FILE] COMPARE RECOGNIZE" );
	    myfree(infile);
	    return;
	    }
	compare_recog(infp);
	fclose(infp);
	myfree(infile);
	if (Timing_flag)
	    show_timing();
	break;

    case KW_PAIRS:		/* FILE COMPARE PAIRS <infile> */
	infile = (char *)strtok8(NULL,whiteSpc);
	if (wanthelp(infile,"COMPARE PAIRS [<file>] (default name is data.pai, default type is .pai)"))
	    break;
	infile = set_filename( infile, "data.pai", ".pai");
	infp = fopen( infile, "r");
	if (infp == (FILE *)NULL)
	    {
	    report_error(FATAL, &Bad_infile, (int *)NULL, (char *)NULL, infile,
						"[FILE] COMPARE PAIRS" );
	    myfree(infile);
	    return;
	    }
				/* cross compare lexical and surface forms */
	compare_pairs(infp,infile);
	fclose(infp);
	myfree(infile);
	if (Timing_flag)
	    show_timing();
	break;

    default:
	report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
						"[FILE] COMPARE", arg);
	break;
    }
}

/*****************************************************************************
 * NAME
 *    do_file
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Finish parsing and execute a FILE command.
 * RETURN VALUE
 *    none
 */
static void do_file()
{
register char *arg;
FILE *infp, *outfp;
int linenum;
int numdots;
RESULT *resp;

arg = (char *)strtok8(NULL,whiteSpc);
switch (lookup( arg, file_tab, n_file, ""))
    {
    case LLNULL:			/* empty command */
	report_error(FATAL, &Miss_keyword, (int *)NULL, (char *)NULL, "FILE" );
	break;

    case LLAMBI:			/* ambiguous command */
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"FILE", arg);
	break;

    case LLHELP:			/* ? */
	fprintf(stderr, "Type HELP FILE for more help\n");
	break;

    case KW_COMPARE:		/* FILE COMPARE ... */
	do_compare();			/* same as COMPARE ... */
	break;

    case KW_GENERATE:		/* FILE GENERATE <infile> [<outfile>] */
	arg = (char *)strtok8(NULL,whiteSpc);
	if (arg == (char *)NULL)
	    {
	    report_error(FATAL, &Miss_infile, (int *)NULL, (char *)NULL,
							"FILE GENERATE");
	    return;
	    }
	if (wanthelp(arg,"FILE GENERATE <infile> [<outfile>] (no default filename or type)"))
	    return;
	infp = fopen( arg, "r");
	if (infp == (FILE *)NULL)
	    {
	    report_error(FATAL, &Bad_infile, (int *)NULL, (char *)NULL, arg,
							"FILE GENERATE");
	    return;
	    }
	arg = (char *)strtok8(NULL,whiteSpc);
	if (arg == (char *)NULL)
	    outfp = stderr;
	else
	    {
	    outfp = fopen( arg, "w" );
	    Fileout_fp = outfp;
	    }
	if (outfp == (FILE *)NULL)
	    {
	    report_error(FATAL, &Bad_outfile, (int *)NULL, (char *)NULL,
						arg, "FILE GENERATE");
	    outfp = stderr;
	    }
	if (Timing_flag)
	    start_timing();
	linenum = 1;
	numdots = 78;
	while ((arg=(char *)getline(infp,&linenum,commentChr)) != (char *)NULL)
	    {
	    arg = (char *)strtok8((unsigned char *)arg, whiteSpc);
	    if (arg != (char *)NULL)
		{
		fprintf(outfp, "%s\n", arg);
		if (Log_fp != (FILE *)NULL)
		    fprintf(Log_fp, "%s\n", arg);

		resp = generator( (unsigned char *)arg,
				  &Lang, Limit_flag, Trace_flag, Log_fp );

		printResults( resp, outfp);
		if (outfp != stderr)
		    {
		    if (numdots-- > 0)
			{
			putc('.', stderr);
#ifdef THINK_C
			fflush(stderr);
#endif
			}
		    else
			{
			fprintf(stderr, "\n.");
			numdots = 77;	/* one dot of 78 already printed */
			}
		    }
		if (Log_fp != (FILE *)NULL)
		    printResults( resp, Log_fp );
		free_result(resp);
		}
	    }
	if (Timing_flag)
	    stop_timing();
	fclose(infp);
	if (outfp != stderr)
	    {
	    if (numdots != 78)
		putc('\n',stderr);	/* newline unless no dots printed */
	    Fileout_fp = (FILE *)NULL;
	    fclose(outfp);
	    }
	else
	    fflush(outfp);
	if (Timing_flag)
	    show_timing();
	break;

    case KW_RECOGNIZE:		/* FILE RECOGNIZE <infile> [<outfile>] */
	arg = (char *)strtok8(NULL,whiteSpc);
	if (arg == (char *)NULL)
	    {
	    report_error(FATAL, &Miss_infile, (int *)NULL, (char *)NULL,
							"FILE RECOGNIZE");
	    return;
	    }
	if (wanthelp(arg,"FILE RECOGNIZE <infile> [<outfile>] (no default filename or type)"))
	    return;
	infp = fopen( arg, "r");
	if (infp == (FILE *)NULL)
	    {
	    report_error(FATAL, &Bad_infile, (int *)NULL, (char *)NULL, arg,
							"FILE RECOGNIZE" );
	    return;
	    }
	arg = (char *)strtok8(NULL,whiteSpc);
	if (arg == (char *)NULL)
	    outfp = stderr;
	else
	    {
	    outfp = fopen( arg, "w" );
	    Fileout_fp = outfp;
	    }
	if (outfp == (FILE *)NULL)
	    {
	    report_error(FATAL, &Bad_outfile,(int *)NULL, (char *)NULL,
						arg, "FILE RECOGNIZE");
	    outfp = stderr;
	    }
	if (Timing_flag)
	    start_timing();
	/*
	 *  for each non-blank line, process first item on the line
	 */
	linenum = 1;
	numdots = 78;
	while ((arg=(char *)getline(infp,&linenum,commentChr)) != (char *)NULL)
	    {
	    arg = (char *)strtok8((unsigned char *)arg, whiteSpc);
	    if (arg != (char *)NULL)
		{
		fprintf(outfp, "%s\n", arg);
		if (Log_fp != (FILE *)NULL)
		    fprintf(Log_fp, "%s\n", arg);

		resp = recognizer( (unsigned char *)arg, &Lang,Limit_flag,Trace_flag,Log_fp );

		printResults( resp, outfp);
		if (outfp != stderr)
		    {
		    if (numdots-- > 0)
			{
			putc('.', stderr);
#ifdef THINK_C
			fflush(stderr);
#endif
			}
		    else
			{
			fprintf(stderr, "\n.");
			numdots = 77;	/* one dot of 78 already printed */
			}
		    }
		if (Log_fp != (FILE *)NULL)
		    printResults( resp, Log_fp );
		free_result(resp);
		}
	    }
	if (Timing_flag)
	    stop_timing();
	fclose(infp);
	if (outfp != stderr)
	    {
	    if (numdots != 78)
		putc('\n',stderr);	/* newline unless no dots printed */
	    fclose(outfp);
	    Fileout_fp = (FILE *)NULL;
	    }
	else
	    fflush(outfp);
	if (Timing_flag)
	    show_timing();
	break;

    default:
	report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
								"FILE", arg);
	break;
    }
}

/*****************************************************************************
 * NAME
 *    do_list
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Finish parsing and execute a LIST command.
 * RETURN VALUE
 *    none
 */
static void do_list()
{
register char *arg;

arg = (char *)strtok8(NULL,whiteSpc);
switch (lookup( arg, list_tab, n_list, ""))
    {
    case LLNULL:			/* empty command */
	report_error(FATAL, &Miss_keyword, (int *)NULL, (char *)NULL, "LIST" );
	break;

    case LLAMBI:			/* ambiguous command */
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"LIST", arg);
	break;

    case LLHELP:			/* ? */
	fprintf(stderr, "Type HELP LIST for more help\n");
	break;

    case KW_LEXICON:		/* LIST LEXICON */
	list_lexicon();
	break;

    case KW_PAIRS:		/* LIST PAIRS */
	list_pairs();
	break;

    case KW_RULES:		/* LIST RULES */
	list_rules();
	break;

    default:
	report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"LIST", arg);
	break;
    }
}

/*****************************************************************************
 * NAME
 *    do_load
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Finish parsing and execute a LOAD command.
 * RETURN VALUE
 *    none
 */
static void do_load()
{
register char *arg;

arg = (char *)strtok8(NULL,whiteSpc);
switch (lookup( arg, load_tab, n_load, ""))
    {
    case LLNULL:			/* empty command */
	report_error(FATAL, &Miss_keyword, (int *)NULL, (char *)NULL,
							"LOAD" );
	break;

    case LLAMBI:			/* ambiguous command */
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"LOAD", arg);
	break;

    case LLHELP:			/* ? */
	fprintf(stderr, "Type HELP LOAD for more help\n");
	break;

    case KW_RULES:			/* LOAD RULES [<filename>] */
	arg = (char *)strtok8((unsigned char *)NULL, whiteSpc);
	if (wanthelp(arg,"LOAD RULES [<file>] (default name is rules.rul, default type is .rul)"))
	    return;
	if (rule_file != (char *)NULL)
	    {
	    myfree(rule_file);
	    rule_file = (char *)NULL;
	    }
	arg = set_filename( arg, "rules.rul", ".rul");
	free_rules(&Lang);		/* erase existing rules */
	if (load_rules( (unsigned char *)arg,&Lang,commentChr) < 0)
	    myfree(arg);		/* error loading the rules */
	else
	    rule_file = (char *)arg;
	break;

    case KW_LEXICON:			/* LOAD LEXICON [<filename>] */
	arg = (char *)strtok8(NULL, whiteSpc);
	if (wanthelp(arg,"LOAD LEXICON [<file>] (default name is lexicon.lex, default type is .lex)"))
	    return;
	if (lex_file != (char *)NULL)
	    {
	    free_lexicons(&Lang);
	    myfree(lex_file);
	    lex_file = (char *)NULL;
	    }
	if (rule_file == (char *)NULL)
	    report_error(FATAL, &Need_rules_first, (int *)NULL, (char *)NULL);
	else
	    {
	    arg = set_filename( arg, "lexicon.lex", ".lex");
	    free_lexicons(&Lang);	/* just for safety if defaulting */
	    if (load_lexicons( (unsigned char *)arg,&Lang,commentChr) < 0)
		myfree(arg);		/* error loading the lexicon */
	    else
		lex_file = arg;
	    }
	break;

    default:
	report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"LOAD", arg);
	break;
    }
}

/*****************************************************************************
 * NAME
 *    do_set
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Finish parsing and execute a SET command.
 * RETURN VALUE
 *    none
 */
static void do_set()
{
register char *arg;
int set_val;

arg = (char *)strtok8(NULL,whiteSpc);
switch (lookup( arg, set_tab, n_set, ""))
    {
    case LLNULL:			/* empty command */
	report_error(FATAL, &Miss_keyword, (int *)NULL, (char *)NULL,
							"SET" );
	break;

    case LLAMBI:			/* ambiguous command */
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"SET", arg);
	break;

    case LLHELP:			/* ? */
	fprintf(stderr, "Type HELP SET for more help\n");
	break;

    case KW_COMMENT:			/* SET COMMENT <char> */
	arg = (char *)strtok8(NULL,whiteSpc);
	if (arg == (char *)NULL)
	    {
	    report_error(FATAL, &Miss_argument, (int *)NULL, (char *)NULL,
							"SET COMMENT");
	    return;
	    }
	if (wanthelp(arg,"Single character used to begin comments in files (use ?? for ?)"))
	    return;
	commentChr = *arg;
	break;

    case KW_LIMIT:		/* SET LIMIT {ON|OFF} */
	arg = (char *)strtok8(NULL,whiteSpc);
	switch (lookup( arg, onoff_tab, n_onoff, ""))
	    {
	    case LLNULL:
		report_error(FATAL, &Miss_keyword, (int *)NULL, (char *)NULL,
							"SET LIMIT");
		break;
	    case LLAMBI:
		report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"SET LIMIT",arg);
		break;
	    case LLHELP:
		return;			/* already gave help */
	    case KW_ON:
		Limit_flag = TRUE;
		break;
	    case KW_OFF:
		Limit_flag = FALSE;
		break;
	    default:
		report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"SET LIMIT",arg);
		break;
	    }
	break;
	

    case KW_RULES:		/* SET RULES {ON|OFF} {ALL|<numbers>} */
	arg = (char *)strtok8(NULL,whiteSpc);
	switch (lookup( arg, onoff_tab, n_onoff, ""))
	    {
	    case LLNULL:
		report_error(FATAL, &Miss_keyword, (int *)NULL, (char *)NULL,
							"SET RULES");
		return;
	    case LLAMBI:
		report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"SET RULES",
									arg );
		return;
	    case LLHELP:
		return;			/* already gave help */
	    case KW_ON:
		set_val = TRUE;
		break;
	    case KW_OFF:
		set_val = FALSE;
		break;
	    default:
		report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"SET RULES",arg);
		return;
	    }
	arg = (char *)strtok8(NULL,whiteSpc);
	if (arg == (char *)NULL)
	    {
	    report_error(FATAL, &Miss_argument, (int *)NULL, (char *)NULL,
							"SET RULES" );
	    return;
	    }
	if (wanthelp(arg, "Either keyword ALL or one or more rule numbers"))
	    return;
	strlower(arg);
	if (	(strcmp(arg,"all") == 0) ||
		(strcmp(arg,"al") == 0) ||
		(strcmp(arg,"a") == 0) )
	    set_rules(NULL,set_val);		/* set all of the rules */
	else if (isascii(*arg) && isdigit(*arg))
	    set_rules( arg, set_val );  /* set the specified rule(s) */
	else
	    report_error(FATAL, &Bad_argument, (int *)NULL, (char *)NULL,
							"SET RULES", arg);
	break;

    case KW_TIMING:		/* SET TIMING {ON|OFF} */
	arg = (char *)strtok8(NULL,whiteSpc);
	switch (lookup( arg, onoff_tab, n_onoff, ""))
	    {
	    case LLNULL:
		report_error(FATAL, &Miss_keyword, (int *)NULL, (char *)NULL,
							"SET TIMING");
		break;
	    case LLAMBI:
		report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"SET TIMING",
									arg );
		break;
	    case LLHELP:
		return;			/* already gave help */
	    case KW_ON:
		Timing_flag = TRUE;
		break;
	    case KW_OFF:
		Timing_flag = FALSE;
		break;
	    default:
		report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"SET TIMING",arg);
		break;
	    }
	break;

    case KW_TRACING:		/* SET TRACING {ON|OFF} */
	arg = (char *)strtok8(NULL,whiteSpc);
	switch (lookup( arg, onoff_tab, n_onoff, NULL) )
	    {
	    case LLNULL:
		report_error(FATAL, &Miss_keyword, (int *)NULL, (char *)NULL,
							"SET TRACING");
		break;
	    case LLAMBI:
		report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"SET TRACING", arg );
		break;
	    case LLHELP:
		fprintf(stderr,
			"Either ON or OFF or a number between 0 and 3\n" );
		break;
	    case KW_ON:
		Trace_flag = 2;		/* ON => trace level 2 */
		break;
	    case KW_OFF:
		Trace_flag = 0;		/* OFF => trace level 0 */
		break;
	    default:
		if (strcmp(arg,"0")==0)
		    Trace_flag = 0;
		else if (strcmp(arg,"1")==0)
		    Trace_flag = 1;
		else if (strcmp(arg,"2")==0)
		    Trace_flag = 2;
		else if (strcmp(arg,"3")==0)
		    Trace_flag = 3;
		else
		    report_error(FATAL, &Bad_argument,(int *)NULL,(char *)NULL,
							"SET TRACING", arg);
		break;
	    }
	break;

    case KW_VERBOSE:		/* SET VERBOSE {ON|OFF} */
	arg = (char *)strtok8(NULL,whiteSpc);
	switch (lookup( arg, onoff_tab, n_onoff, ""))
	    {
	    case LLNULL:
		report_error(FATAL, &Miss_keyword, (int *)NULL, (char *)NULL,
							"SET VERBOSE");
		break;
	    case LLAMBI:
		report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"SET VERBOSE", arg);
		break;
	    case LLHELP:
		return;			/* already gave help */
	    case KW_ON:
		Verbose_flag = TRUE;
		break;
	    case KW_OFF:
		Verbose_flag = FALSE;
		break;
	    default:
		report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"SET VERBOSE",arg);
		break;
	    }
	break;

    default:
	report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"SET", arg);
	break;
    }
}

/*****************************************************************************
 * NAME
 *    do_show
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Finish parsing and execute a SHOW command.
 * RETURN VALUE
 *    none
 */
static void do_show()
{
register char *arg;
int k;

arg = (char *)strtok8(NULL,whiteSpc);
switch (lookup( arg, show_tab, n_show, ""))
    {
    case LLNULL:			/* empty */
	show_status();			/* SHOW == SHOW STATUS == STATUS */
	break;

    case LLAMBI:			/* ambiguous */
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"SHOW", arg);
	break;

    case LLHELP:			/* ? */
	fprintf(stderr, "Type HELP SHOW for more help\n");
	break;

    case KW_LEXICON:		/* SHOW LEXICON <name> */
	arg = (char *)strtok8(NULL,whiteSpc);
	if (wanthelp(arg,"SHOW LEXICON section_name (no default)"))
	    return;
	show_lexicon( (unsigned char *)arg);
	break;

    case KW_RULES:		/* SHOW RULE <number> */
	arg = (char *)strtok8(NULL,whiteSpc);
	if (wanthelp(arg,"SHOW RULE rule_number (no default)"))
	    return;
	if (arg == (char *)NULL)
	    report_error(FATAL, &Miss_keyword, (int *)NULL, (char *)NULL,
							"SHOW RULE");
	else if (   !isascii(*arg) || !isdigit(*arg) ||
		    ((k = atoi(arg)) <= 0) || (k > Lang.num_rules) )
	    report_error(FATAL, &Bad_argument, (int *)NULL, (char *)NULL,
							"SHOW RULE", arg);
	else
	    show_rules( k );
	break;

    case KW_STATUS:		/* SHOW STATUS */
	show_status();
	break;

    default:
	report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"SHOW", arg);
	break;
    }
}

/*****************************************************************************
 * NAME
 *    do_hlp_status
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Display information for the HELP STATUS command.
 * RETURN VALUE
 *    none
 */
static void do_hlp_status()
{
fputs("\n\
STATUS displays the names of the current rules, lexicon, and log files, and\n\
the values of the parameters established by the SET command.\n\
\n\
SHOW STATUS, or SHOW (by itself), are synonyms for STATUS.\n\
\n", stderr);
}

/*****************************************************************************
 * NAME
 *    do_hlp_system
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Display information for the HELP SYSTEM command.
 * RETURN VALUE
 *    none
 */
static void do_hlp_system()
{
#ifdef THINK_C
fputs("\
NOT IMPLEMENTED FOR MACINTOSH.\n\
", stderr);
#else
fputs("\
SYSTEM [<command>] allows the user to execute an operating system command\n\
(such as listing the files in the current directory) from within PC-KIMMO.\n\
If no system-level command is given on the line with the SYSTEM command,\n\
then PC-KIMMO is pushed into the background and a new system command processor\n\
(shell) is started.  Control is usually returned to PC-KIMMO in this case by\n\
typing EXIT as the operating system command.\n\
\n\
! (exclamation point) is a synonym for SYSTEM.\n\
", stderr);
#endif /*THINK_C*/
}

/*****************************************************************************
 * NAME
 *    do_hlp_compare
 * ARGUMENTS
 *    key - pointer to next keyword for command, or NULL if none
 * DESCRIPTION
 *    Display information for the HELP COMPARE command.
 * RETURN VALUE
 *    none
 */
static void do_hlp_compare(key)
unsigned char *key;
{
switch (lookup( (char *)key, compare_tab, n_compare, ""))
    {
    case LLNULL:			/* empty command */
	fputs("\
[FILE] COMPARE GENERATE <file>  produces and compares surface forms from a file\n\
[FILE] COMPARE RECOGNIZE <file> produces and compares lexical forms from a file\n\
[FILE] COMPARE PAIRS <file>     produces and compares both surface and lexical\n\
                                forms from a file\n\
\n\
COMPARE is a synonym for FILE COMPARE.\n\
\n\
Type HELP [FILE] COMPARE GENERATE, HELP [FILE] COMPARE RECOGNIZE, or\n\
HELP [FILE] COMPARE PAIRS for more information.\n\
", stderr);
	break;

    case LLAMBI:			/* ambiguous command */
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
						"HELP [FILE] COMPARE", key);
	break;

    case LLHELP:			/* ? */
	fputs("\
HELP [FILE] COMPARE without an argument displays a list of the [FILE] COMPARE\n\
commands\n", stderr);
	break;

    case KW_GENERATE:		/* FILE COMPARE GENERATE <infile> */
	fputs("\
[FILE] COMPARE GENERATE <file> reads lexical and surface forms from the\n\
specified file.  After reading a lexical form, PC-KIMMO generates the\n\
corresponding surface form(s) and compares the result to the surface form(s)\n\
read from the file.  If VERBOSE is ON, then each form from the file is echoed\n\
on the screen with a message indicating whether or not the surface forms\n\
generated by PC-KIMMO and read from the file are in agreement.  If VERBOSE\n\
", stderr);
	fputs("\
is OFF, then only the disagreements in surface form are displayed fully.\n\
Each result which agrees is indicated by a single dot written to the screen.\n\
\n\
The default filetype extension for FILE COMPARE GENERATE is \".gen\", and the\n\
default filename is \"data.gen\" (without the quotation marks, of course).\n\
\n\
COMPARE GENERATE <file> is a synonym for FILE COMPARE GENERATE <file>.\n\
", stderr);
	break;

    case KW_RECOGNIZE:		/* FILE COMPARE RECOGNIZE <infile> */
	fputs("\
[FILE] COMPARE RECOGNIZE <file> reads surface and lexical forms from the\n\
specified file.  After reading a surface form, PC-KIMMO produces any\n\
corresponding lexical form(s) and compares the result(s) to the lexical\n\
form(s) read from the file.  If VERBOSE is ON, then each form from the file\n\
is echoed on the screen with a message indicating whether or not the lexical\n\
forms produced by PC-KIMMO and read from the file are in agreement.  If\n\
", stderr);
	fputs("\
VERBOSE is OFF, then each result which agrees is indicated by a single dot\n\
written to the screen, and only disagreements in lexical forms are displayed\n\
fully.\n\
\n\
The default filetype extension for FILE COMPARE RECOGNIZE is \".rec\", and\n\
the default filename is \"data.rec\" (without the quotation marks, of course).\n\
\n\
COMPARE RECOGNIZE <file> is a synonym for FILE COMPARE RECOGNIZE <file>.\n\
", stderr);
	break;

    case KW_PAIRS:		/* FILE COMPARE PAIRS <infile> */
	fputs("\
[FILE] COMPARE PAIRS <file> reads pairs of surface and lexical forms from the\n\
specified file.  After reading a lexical form, PC-KIMMO produces any\n\
corresponding surface form(s) and compares the result(s) to the surface\n\
form read from the file.  For each surface form, PC-KIMMO also produces any\n\
corresponding lexical form(s) and compares the result to the lexical form\n\
read from the file.  If VERBOSE is ON, then each form from the file is echoed\n\
", stderr);
	fputs("\
on the screen with a message indicating whether or not the forms produced by\n\
PC-KIMMO and read from the file are in agreement.  If VERBOSE is OFF, then\n\
each result which agrees is indicated by a single dot written to the screen,\n\
and only disagreements in lexical forms are displayed fully.\n\
\n\
The default filetype extension for FILE COMPARE PAIRS is \".pai\", and the\n\
default filename is \"data.pai\" (without the quotation marks, of course).\n\
\n\
COMPARE PAIRS <file> is a synonym for FILE COMPARE PAIRS <file>.\n\
", stderr);
	break;

    default:
	report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
						"HELP [FILE] COMPARE", key);
	break;
    }
}

/*****************************************************************************
 * NAME
 *    do_hlp_file
 * ARGUMENTS
 *    key - next keyword for HELP FILE command
 * DESCRIPTION
 *    Display information for the HELP FILE command
 * RETURN VALUE
 *    none
 */
static void do_hlp_file(key)
unsigned char *key;
{
switch (lookup( (char*)key, file_tab, n_file, ""))
    {
    case LLNULL:			/* empty command */
	fputs("\
[FILE] COMPARE {GENERATE|RECOGNIZE|PAIRS} <infile>\n\
                read test data from a file, comparing computed results\n\
FILE GENERATE <infile> [<outfile>]\n\
                read lexical forms from a file, computing surface forms\n\
FILE RECOGNIZE <infile> [<outfile>]\n\
                read surface forms from a file, computing lexical forms\n\
\n\
Type HELP FILE COMPARE, HELP FILE GENERATE, OR HELP FILE RECOGNIZE for more\n\
information.\n\
", stderr);
	break;

    case LLAMBI:			/* ambiguous command */
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"HELP FILE", key);
	break;

    case LLHELP:			/* ? */
	fputs("\
HELP FILE without an argument displays a list of the FILE commands\n", stderr);
	break;

    case KW_COMPARE:		/* FILE COMPARE ... */
	do_hlp_compare(strtok8(NULL,whiteSpc));	/* same as COMPARE ... */
	break;

    case KW_GENERATE:		/* FILE GENERATE <infile> [<outfile>] */
	fputs("\
FILE GENERATE <infile> [<outfile>] reads lexical forms from the specified\n\
intput file and writes the corrsponding computed surface forms either to\n\
the screen or to an optionally specified output file.\n\
\n\
This command behaves the same as GENERATE except that input comes from a file\n\
rather than the keyboard, and output may go to a file rather than the screen.\n\
", stderr);
	break;

    case KW_RECOGNIZE:		/* FILE RECOGNIZE <infile> [<outfile>] */
	fputs("\
FILE RECOGNIZE <infile> [<outfile>] reads surface forms from the specified\n\
intput file and writes the corrsponding computed lexical forms either to\n\
the screen or to an optionally specified output file.\n\
\n\
This command behaves the same as RECOGNIZE except that input comes from a\n\
file rather than the keyboard, and output may go to a file rather than the\n\
screen.\n\
", stderr);
	break;

    default:
	report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"HELP FILE", key);
	break;
    }
}

/*****************************************************************************
 * NAME
 *    do_hlp_list
 * ARGUMENTS
 *    key - next keyword for HELP LIST command
 * DESCRIPTION
 *    Display information for the HELP LIST command.
 * RETURN VALUE
 *    none
 */
static void do_hlp_list(key)
unsigned char *key;
{
switch (lookup( (char*)key, list_tab, n_list, ""))
    {
    case LLNULL:			/* empty command */
	fputs("\
LIST LEXICON  displays the names of all the (sub)lexicons currently loaded\n\
LIST RULES    displays the name and status (ON or OFF) of each of the rules\n\
              currently loaded\n\
LIST PAIRS    displays all the feasible pairs for the current set of active\n\
              rules\n\
\n\
Type HELP LIST LEXICON, HELP LIST RULES, or HELP LIST PAIRS for more\n\
information.\n\
", stderr);
	break;

    case LLAMBI:			/* ambiguous command */
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"HELP LIST", key);
	break;

    case LLHELP:			/* ? */
	fputs("\
HELP LIST without an argument displays a list of the LIST commands\n", stderr);
	break;

    case KW_LEXICON:		/* LIST LEXICON */
	fputs("\
LIST LEXICON displays the names of all the current (sub)lexicons.  The order\n\
of presentation is the order in which they are referenced in the ALTERNATIONS\n\
declarations.\n\
", stderr);
	break;

    case KW_PAIRS:		/* LIST PAIRS */
	fputs("\
LIST PAIRS displays all the feasible pairs associated with the current set\n\
of active rules.  The feasible pairs are displayed as pairs of lines, with\n\
the lexical characters shown above the corresponding surface characters.\n\
", stderr);
	break;

    case KW_RULES:		/* LIST RULES */
	fputs("\
LIST RULES displays the names of the current rules, preceded by the number\n\
of the rule (used by the SET RULES command) and an indication of whether the\n\
rule is ON or OFF.\n\
", stderr);
	break;

    default:
	report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"HELP LIST", key);
	break;
    }
}

/*****************************************************************************
 * NAME
 *    do_hlp_load
 * ARGUMENTS
 *    key - next keyword for HELP LOAD command
 * DESCRIPTION
 *    Display information for the HELP LOAD command.
 * RETURN VALUE
 *    none
 */
static void do_hlp_load(key)
unsigned char *key;
{
switch (lookup( (char*)key, load_tab, n_load, ""))
    {
    case LLNULL:			/* empty command */
	fputs("\
LOAD RULES [<file>]    loads a new rules file\n\
LOAD LEXICON [<file>]  loads a new lexicon file (rules must be loaded first)\n\
\n\
L is a sufficient synonym for LOAD.\n\
\n\
Type HELP LOAD RULES or HELP LOAD LEXICON for more information.\n\
", stderr);
	break;

    case LLAMBI:			/* ambiguous command */
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"HELP LOAD", key);
	break;

    case LLHELP:			/* ? */
	fputs("\
HELP LOAD without an argument displays a list of the LOAD commands\n", stderr);
	break;

    case KW_RULES:			/* LOAD RULES [<filename>] */
	fputs("\
LOAD RULES [<file>] erases any existing rules and reads a new set of rules\n\
from the specified file.\n\
\n\
The default filetype extension for LOAD RULES is \".rul\", and the\n\
default filename is \"rules.rul\" (without the quotation marks, of course).\n\
\n\
A rules file can also be loaded by using the -r command line option when\n\
starting PC-KIMMO.\n\
\n\
L RULES [<file>] is a synonym for LOAD RULES [<file>].\n\
", stderr);
	break;

    case KW_LEXICON:			/* LOAD LEXICON [<filename>] */
	fputs("\
LOAD LEXICON [<file>] erases any existing lexicon information and reads a new\n\
lexicon from the specified file.  A rules file must be loaded before a lexicon\n\
file can be loaded.\n\
\n\
The default filetype extension for LOAD LEXICON is \".lex\", and the\n\
default filename is \"lexicon.lex\" (without the quotation marks, of course).\n\
\n\
A lexicon file can also be loaded by using the -l command line option when\n\
starting PC-KIMMO.\n\
\n\
L LEXICON [<file>] is a synonym for LOAD LEXICON [<file>].\n\
", stderr);
	break;

    default:
	report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"HELP LOAD", key);
	break;
    }
}

/*****************************************************************************
 * NAME
 *    do_hlp_set
 * ARGUMENTS
 *    key - next keyword for HELP SET command
 * DESCRIPTION
 *    Display information for the HELP SET command.
 * RETURN VALUE
 *    none
 */
static void do_hlp_set(key)
unsigned char *key;
{
switch (lookup( (char*)key, set_tab, n_set, ""))
    {
    case LLNULL:			/* empty command */
	fputs("\
SET COMMENT <char>\n\
                selects the comment character for reading from files\n\
SET LIMIT {ON|OFF}\n\
                turns limit output mode ON or OFF\n\
SET RULES {ON|OFF} {ALL|<number(s)}\n\
                turns selected rules ON or OFF\n\
SET TIMING {ON|OFF}\n\
                turns timing mode ON or OFF\n\
SET TRACING {<level>|ON|OFF}\n\
                sets the tracing level (ON = 2, OFF = 0)\n\
SET VERBOSE {ON|OFF}\n\
                turns verbose comparison mode ON or OFF\n\
", stderr);
	fputs("\
\n\
Type HELP SET COMMENT, HELP SET LIMIT, HELP SET RULES, HELP SET TIMING,\n\
HELP SET TRACING, or HELP SET VERBOSE for more information.\n\
", stderr);
	break;

    case LLAMBI:			/* ambiguous command */
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"HELP SET", key);
	break;

    case LLHELP:			/* ? */
	fputs("\
HELP SET without an argument displays a list of the SET commands\n", stderr);
	break;

    case KW_COMMENT:			/* SET COMMENT <char> */
	fputs("\
SET COMMENT <char> sets the comment character for reading from files to the\n\
specified character.  The default value is ; (semicolon).\n\
\n\
The comment character can also be set by using the -c command line option\n\
when starting PC-KIMMO.\n\
", stderr);
	break;

    case KW_LIMIT:			/* SET LIMIT {ON|OFF} */
	fputs("\
SET LIMIT ON limits both the RECOGNIZE and GENERATE commands to a single\n\
result.  SET LIMIT OFF allows both the RECOGNIZE and GENERATE commands to\n\
produce multiple results.  (OFF is the default mode.)\n\
\n\
The LIMIT value affects the FILE RECOGNIZE, FILE GENERATE, and FILE COMPARE\n\
commands as well as the GENERATE and RECOGNIZE commands.\n\
", stderr);
	break;

    case KW_RULES:		/* SET RULES {ON|OFF} {ALL|<numbers>} */
	fputs("\
SET RULES ON ALL turns on all the rules.\n\
SET RULES OFF ALL turns off all the rules.\n\
\n\
SET RULES ON <number>... turns on the rule(s) selected by the number(s).\n\
SET RULES OFF <number>... turns off the rule(s) selected by the number(s).\n\
The numbers to use for the second form of the SET RULES command are those\n\
shown by the LIST RULES command.\n\
", stderr);
	break;

    case KW_TIMING:		/* SET TIMING {ON|OFF} */
	fputs("\
SET TIMING ON turns on timing mode.  SET TIMING OFF turns off timing mode.\n\
If timing mode is ON, then the elapsed time required to process either a\n\
RECOGNIZE or a GENERATE command is displayed when the command finishes.\n\
If timing mode is OFF, then the elapsed time is not shown.\n\
\n\
Timing affects the FILE RECOGNIZE, FILE GENERATE, and FILE COMPARE commands\n\
as well, but the timing is for the entire file, not for individual items.\n\
", stderr);
	break;

    case KW_TRACING:		/* SET TRACING {ON|OFF} */
	fputs("\
SET TRACING {<level>|ON|OFF} establishes the tracing level, which can be any\n\
integer value between 0 and 3.  SET TRACING ON sets the tracing level to 2.\n\
SET TRACING OFF sets the tracing level to 0.  If the tracing level is nonzero,\n\
then either a RECOGNIZE or a GENERATE command causes a step-by-step trace of\n\
the operation of the algorithm to be displayed on the screen (and written to\n\
the log file, if one has been opened).  If the tracing level is zero, then\n\
only the final results are displayed.\n\
", stderr);
	fputs("\n\
Tracing affects the FILE RECOGNIZE, FILE GENERATE, and FILE COMPARE commands\n\
as well.  Be aware that tracing can cause voluminous output, especially at\n\
level 3!\n\
", stderr);
	break;

    case KW_VERBOSE:		/* SET VERBOSE {ON|OFF} */
	fputs("\
SET VERBOSE ON turns on verbose mode.  SET VERBOSE OFF turns off verbose mode.\n\
If verbose mode is ON, then FILE COMPARE commands display on the screen both\n\
the surface and lexical forms read from the input file while processing the\n\
file.  If verbose mode is OFF, then only errors display the forms read from\n\
the file.  Good data is indicated by a single dot written to the screen.\n\
", stderr);
	break;

    default:
	report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"HELP SET", key);
	break;
    }
}

/*****************************************************************************
 * NAME
 *    do_hlp_show
 * ARGUMENTS
 *    key - next keyword for HELP SHOW command
 * DESCRIPTION
 *    Display information for the HELP SHOW command.
 * RETURN VALUE
 *    none
 */
static void do_hlp_show(key)
unsigned char *key;
{
switch (lookup( (char*)key, show_tab, n_show, ""))
    {
    case LLNULL:			/* empty */
	fputs("\
SHOW LEXICON <name>\n\
                displays the contents of the specified (sub)lexicon\n\
SHOW RULE <number>\n\
                displays the state of the specified rule\n\
SHOW STATUS     is a synonym for STATUS\n\
\n\
SHOW            without any argument is also a synonym for STATUS\n\
\n\
Type HELP SHOW LEXICON, HELP SHOW RULE, or HELP [SHOW] STATUS for more\n\
information.\n\
", stderr);
	break;

    case LLAMBI:			/* ambiguous */
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"HELP SHOW", key);
	break;

    case LLHELP:			/* ? */
	fputs("\
HELP SHOW without an argument displays a list of the SHOW commands\n", stderr);
	break;

    case KW_LEXICON:		/* SHOW LEXICON <name> */
	fputs("\
SHOW LEXICON <name> displays the contents of the indicated (sub)lexicon.\n\
For each lexical item, the lexical item is shown first followed by the\n\
lexical gloss enclosed in quotation marks.  The set of valid continuation\n\
(sub)lexicons is shown last on the line, enclosed in curly braces.\n\
", stderr);
	break;

    case KW_RULES:		/* SHOW RULE <number> */
	fputs("\
SHOW RULE <number> displays the specified rule.  First, its number,\n\
status (ON or OFF), and name are shown as in the LIST RULES command.\n\
If the rules is ON, then and feasible pairs associated with each column\n\
of the state table are displayed.  Each feasible pair is shown in the\n\
form Lexical-character:Surface-character.\n\
", stderr);
	break;

    case KW_STATUS:		/* SHOW STATUS */
	do_hlp_status();
	break;

    default:
	report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"HELP SHOW", key);
	break;
    }
}

/*****************************************************************************
 * NAME
 *    do_help
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Finish parsing and execute a HELP command.
 * RETURN VALUE
 *    none
 */
static void do_help()
{
char *arg;

putc('\n', stderr);
arg = (char *)strtok8(NULL,whiteSpc);
switch (lookup( arg, cmd_tab, n_cmd, ""))
    {
    case LLNULL:		/* barebones HELP command */
#ifdef THINK_C
	fputs("\
CD         change the current directory\n\
", stderr);
#endif
	fputs("\
CLOSE      close the current log file\n\
COMPARE    compare prepared data from a file to results of processing data\n\
", stderr);
#ifndef THINK_C
	fputs("\
EDIT       edit a file\n\
", stderr);
#endif
	fputs("\
EXIT       exit PC-KIMMO (same as QUIT)\n\
FILE       process data from a file\n\
GENERATE   process lexical form to produce surface form(s)\n\
HELP       display this list of commands\n\
LIST       list rule names, feasible pairs, or sublexicon names\n\
LOAD       load rules or a lexicon from a file\n\
LOG        open a log file\n\
", stderr);
	fputs("\
NEW        start a new language, erasing existing rules and lexicon\n\
QUIT       quit PC-KIMMO (same as EXIT)\n\
RECOGNIZE  process surface form to produce lexical form(s)\n\
SET        set processing flags\n\
SHOW       show status, rule, or sublexicon\n\
STATUS     show current files and flag settings\n\
", stderr);
#ifndef THINK_C
	fputs("\
SYSTEM     execute an operating system command or program\n\
", stderr);
#endif
	fputs("\
TAKE       take PC-KIMMO commands from a file\n\
\n\
For more information, type HELP with a specific command name.\n\
", stderr);
	break;

    case LLAMBI:
	report_error(FATAL, &Ambig_keyword, (int *)NULL, (char *)NULL,
							"HELP", arg);
	break;

    case LLHELP:
	fputs("HELP without an argument displays a list of the commands\n",
			stderr );
	break;

    case KW_CLOSE:			/* CLOSE */
	fputs("\
CLOSE closes the log file opened by a LOG command.\n\
", stderr);
	break;

    case KW_COMPARE:			/* COMPARE ... */
	do_hlp_compare(strtok8(NULL,whiteSpc));
	break;

    case KW_EDIT:			/* EDIT <file> */
#ifdef MSDOS
	fputs("\
EDIT <file> attempts to edit a file using the program indicated by the\n\
environment variable EDITOR.  If this environment variable is not defined,\n\
then EDLIN is used to edit the file.\n\
", stderr);
#endif
#ifdef UNIX
	fputs("\
EDIT <file> attempts to edit a file using the program indicated by the\n\
environment variable EDITOR.  If this environment variable is not defined,\n\
then vi is used to edit the file.\n\
", stderr);
#endif
#ifdef THINK_C
	fputs("\
NOT IMPLEMENTED FOR MACINTOSH.\n\
", stderr);
#endif /*THINK_C*/
	break;

#ifdef THINK_C
	case KW_CD:				/* CD ... */
	fputs("\
CD <pathname> changes the current directory.  You can give a full path\n\
starting with the name of your hard disk, a path starting with : which\n\
means the current folder, or one starting :: which means the folder\n\
containing the current one (and so on).  Spaces in names are not\n\
permitted.\n\
", stderr);
	break;
#endif

    case KW_FILE:			/* FILE ... */
	do_hlp_file(strtok8(NULL,whiteSpc));
	break;

    case KW_GENERATE:			/* GENERATE [<form>] */
	fputs("\
GENERATE [<lexical-form>] attempts to produce a surface form from a lexical\n\
form provided by the user.  If the command is typed without a form, then\n\
PC-KIMMO prompts the user for lexical forms with a special generator prompt,\n\
processing each form and asking for another until the user responds to the\n\
generator prompt with an empty line.\n\
", stderr);
	break;

    case KW_HELP:			/* HELP ... */
	fputs("\
HELP [<command-name>] displays a description of a command.  If HELP is typed\n\
by itself, PC-KIMMO displays a list of commands with short descriptions of\n\
each command.\n\
", stderr);
	break;

    case KW_LIST:			/* LIST ... */
	do_hlp_list(strtok8(NULL,whiteSpc));
	break;

    case KW_LOAD:			/* LOAD ... */
	do_hlp_load(strtok8(NULL,whiteSpc));
	break;

    case KW_LOG:			/* LOG [<filename>] */
	fputs("\
LOG [<file>] opens a log file.  Each item processed by a GENERATE, RECOGNIZE,\n\
FILE GENERATE, FILE RECOGNIZE, or [FILE] COMPARE command is recorded in the\n\
log file as well as being displayed on the screen.\n\
\n", stderr);
	fputs("\
If a filename is given on the same line as the LOG command, then that file\n\
is used for the log file.  Any previously existing file with the same name\n\
will be overwritten.  If no filename is provided, then the file pckimmo.log\n\
in the current directory is used for the log file.\n\
\n", stderr);
	fputs("\
Use CLOSE to stop recording in a log file.  If a LOG command is given when\n\
a log file is already open, then the earlier log file is closed before the\n\
new log file is opened.\n\
", stderr);
	break;

    case KW_NEW:			/* NEW */
	fputs("\
NEW erases all rules and lexicon information, allowing the user to prepare\n\
to load information for a new language.  Strictly speaking, it is not needed\n\
since the LOAD RULES command erases all previously existing rules, and the\n\
LOAD LEXICON command erases any previously existing lexicon.\n\
", stderr);
	break;

    case KW_QUIT:			/* QUIT or EXIT */
	fputs("\
Either EXIT or QUIT stops PC-KIMMO.\n\
", stderr);
	break;

    case KW_RECOGNIZE:			/* RECOGNIZE [<form>] */
	fputs("\
RECOGNIZE [<surface-form>] attempts to produce lexical forms from a surface\n\
form provided by the user.  If the command is typed without a form, then\n\
PC-KIMMO prompts the user for surface forms with a special recognizer prompt,\n\
processing each form and asking for another until the user responds to the\n\
recognizer prompt with an empty line.\n\
", stderr);
	break;

    case KW_SET:			/* SET ... */
	do_hlp_set(strtok8(NULL,whiteSpc));
	break;

    case KW_SHOW:			/* SHOW ... */
	do_hlp_show(strtok8(NULL,whiteSpc));
	break;

    case KW_STATUS:			/* STATUS */
	do_hlp_status();
	break;

    case KW_SYSTEM:			/* SYSTEM [<command>] */
	do_hlp_system();
	break;

    case KW_TAKE:			/* TAKE <filename> */
	fputs("\
TAKE [<file>] redirects command input to the specified file.\n\
\n\
The default filetype extension for TAKE is \".tak\", and the default\n\
filename is \"pckimmo.tak\" (without the quotation marks, of course).\n\
\n\
TAKE files can be nested three deep.  That is, the user types TAKE file1,\n\
file1 contains the command TAKE file2, and file2 has the command TAKE file3.\n\
It would be an error for file3 to contain a TAKE command.  This should not\n\
prove to be a serious limitation.\n\
", stderr);
	fputs("\n\
A TAKE file can also be specified by using the -t command line option when\n\
starting PC-KIMMO.\n\
", stderr);
	break;

    default:
	if (*arg == '!')
	    do_hlp_system();
	else
	    report_error(FATAL, &Bad_keyword, (int *)NULL, (char *)NULL,
							"HELP", arg);
	break;
    }
putc('\n', stderr);
}

/*****************************************************************************
 * NAME
 *    do_cmd
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Parse and execute one command.
 * RETURN VALUE
 *    zero for QUIT or EXIT command, nonzero for everything else
 */
static int do_cmd()
{
register char *cmd, *arg;
static char cmdbuf[MAXLINELEN+1];
char *savcmd;
/*
 *  Read the next command from the appropriate place
 */
while ((tlevel >= 0) && feof(tfile[tlevel]))
    {
    fclose(tfile[tlevel]);		/* if end of take file, close it */
    --tlevel;				/*    and forget about it. */
    }
if (tlevel > -1)
    {
    cmd = (char *)getline( tfile[tlevel], &tlinenum[tlevel], commentChr );
    if (cmd == (char *)NULL)
	return(1);		/* catch the EOF next time around */
    fprintf(stderr, "%s%s\n", prompt, cmd);
    }
else
    {
    fputs( prompt, stderr);
    cgetline((unsigned char *)cmdbuf,MAXLINELEN);	/* get command line */
    cmd = cmdbuf;
    }
if (*cmd == '\0')
    return(1);				/* ignore empty commands */
/*
 *  parse the first keyword of the command
 */
savcmd = (char *)mystrdup(cmd);
arg = (char *)strtok8((unsigned char *)cmd,whiteSpc);
switch (lookup( arg, cmd_tab, n_cmd, "Command"))
    {
    case LLNULL:			/* empty command */
	break;				/* (just ignore it) */

    case LLAMBI:			/* ambiguous command */
	report_error(FATAL, &Ambig_command, (int *)NULL, (char *)NULL,
							arg);
	break;

    case LLHELP:			/* ? - list of commands displayed */
	fprintf(stderr, "Type HELP for more help\n");
	break;
	
#ifdef THINK_C
	case KW_CD:				/* CD, change directory */
	do_cd( (char *)strtok8((unsigned char *)NULL,whiteSpc) );
	break;
#endif

    case KW_CLOSE:			/* CLOSE */
	do_close();
	break;

    case KW_COMPARE:			/* COMPARE ... */
	do_compare();
	break;

#ifndef THINK_C
    case KW_EDIT:			/* EDIT <file> */
	do_edit( (char *)strtok8((unsigned char *)NULL,whiteSpc) );
	break;
#endif /*THINK_C*/

    case KW_FILE:			/* FILE ... */
	do_file();
	break;

    case KW_GENERATE:			/* GENERATE [<form>] */
	do_generate( strtok8((unsigned char *)NULL,whiteSpc) );
	break;

    case KW_HELP:			/* HELP ... */
	do_help();
	break;

    case KW_LIST:			/* LIST ... */
	do_list();
	break;

    case KW_LOAD:			/* LOAD ... */
	do_load();
	break;

    case KW_LOG:			/* LOG [<filename>] */
	do_log( (char *)strtok8((unsigned char *)NULL,whiteSpc) );
	break;

   case KW_NEW:			/* NEW */
	do_new();
	break;

    case KW_QUIT:			/* QUIT or EXIT */
	myfree(savcmd);
	return( 0 );

    case KW_RECOGNIZE:			/* RECOGNIZE [<form>] */
	do_recognize( strtok8((unsigned char *)NULL,whiteSpc) );
	break;

    case KW_SET:			/* SET ... */
	do_set();
	break;

    case KW_SHOW:			/* SHOW ... */
	do_show();
	break;

    case KW_STATUS:			/* STATUS */
	show_status();			/* STATUS == SHOW STATUS */
	break;

#ifndef THINK_C
    case KW_SYSTEM:			/* SYSTEM [<command>] */
	/*
	 *  retrieve the saved copy of the original command string
	 *    and skip past the initial keyword
	 *  try to execute the rest of the command line as a system command
	 */
#ifdef UNIX
	if (wanthelp((char *)strtok8(NULL,whiteSpc),"\
SYSTEM [<command>] (default is subshell, type EXIT to return to PC-KIMMO)"))
#endif
#ifdef MSDOS
	if (wanthelp((char *)strtok8(NULL,whiteSpc),"\
SYSTEM [<command>] (default is DOS shell, type EXIT to return to PC-KIMMO)"))
#endif
	    return(1);
	arg = savcmd;
	while (isascii(*arg) && isspace(*arg))
	    ++arg;
	while ( !isascii(*arg) || !isspace(*arg) )
	    {
	    if (*arg == NUL)
		break;
	    ++arg;
	    }
	do_system( (unsigned char *)arg );
	break;
#endif /*THINK_C*/

    case KW_TAKE:			/* TAKE <filename> */
	arg = (char *)strtok8(NULL,whiteSpc);
	if (wanthelp(arg,"TAKE [<file>] (default name is pckimmo.tak, default type is .tak)"))
	    return(1);
	arg = set_filename( arg, "pckimmo.tak", ".tak");
	if (tlevel >= MAXTAKE-1)
	    report_error(FATAL, &Too_deep_TAKE, (int *)NULL, (char *)NULL);
	else if ((tfile[++tlevel] = fopen(arg,"r")) == (FILE *)NULL)
	    {
	    report_error(FATAL, &Bad_infile, (int *)NULL, (char *)NULL,
							arg, "TAKE");
	    tlevel--;
	    }
	else
	    tlinenum[tlevel] = 1;
	break;

    default:				/* invalid command */
#ifndef THINK_C
	if (*savcmd == '!')		/* treat ! same as SYSTEM, except */
	    {				/* that no space is needed before */
	    do_system((unsigned char *)savcmd+1);   /* the system command */
	    break;
	    }
#endif  /*THINK_C*/    
	if (tlevel > -1)
	    {				/* If in take file, abort it */
	    report_error(FATAL, &Error_in_TAKE, (int *)NULL, (char *)NULL,
							arg);
	    fclose(tfile[tlevel]);
	    tlevel--;
	    }
	else
	    report_error(FATAL, &Bad_command, (int *)NULL, (char *)NULL,
							arg);
	break;
    }
myfree(savcmd);
return(1);			/* tell caller to keep going */
}

/*****************************************************************************
 * NAME
 *    user_cmd
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Interactively parse commands from the user.
 * RETURN VALUE
 *    none
 */
void user_cmd()
{
fprintf(stderr, "Type ? for help\n");

Lang.alphabet = (unsigned char *)NULL;	/* no alphabet yet */
Lang.subsets = (SUBSET *)NULL;		/* no subsets yet */
Lang.numsubsets = 0;
Lang.automata = (RULE *)NULL;		/* no rules yet */
Lang.num_rules = 0;
Lang.lex_pair = (unsigned char *)NULL;	/* no feasible pairs yet */
Lang.surf_pair = (unsigned char *)NULL;
Lang.num_pairs = 0;
Lang.alterns = (ALTERNATION *)NULL;	/* no alternations yet */
Lang.num_alterns = 0;
Lang.lex_sections = (LEXICON *)NULL;	/* no lexicon sections yet */
Lang.num_lex_sections = 0;
/*
 *  handle the command line arguments
 */
rule_file = set_filename( rule_file, (char *)NULL, ".rul");
if (rule_file != (char *)NULL)			/* -r <rulefile> */
    {
    fprintf(stderr, "%sLOAD RULES %s\n", prompt, rule_file);
    if (load_rules( (unsigned char *)rule_file,&Lang,commentChr) < 0)
	{
	myfree(rule_file);
	rule_file = (char *)NULL;	/* error loading the rules */
	}
    }
lex_file = (char *)set_filename( lex_file, (char *)NULL, ".lex" );
if (lex_file != (char *)NULL)			/* -l <lexfile> */
    {
    fprintf(stderr, "%sLOAD LEXICON %s\n", prompt, lex_file);
    if (rule_file == (char *)NULL)
	{
	report_error(FATAL, &Need_rules_first, (int *)NULL, (char *)NULL);
	myfree(lex_file);
	lex_file = (char *)NULL;
	}
    else if (load_lexicons( (unsigned char *)lex_file,&Lang,commentChr) < 0)
	{
	myfree(lex_file);		/* error loading the lexicon */
	lex_file = (char *)NULL;
	}
    }
cmd_file = set_filename( cmd_file, (char *)NULL, ".tak");
if (cmd_file != (char *)NULL)			/* -t <cmdfile> */
    {
    if ((tfile[++tlevel] = fopen(cmd_file,"r")) == (FILE *)NULL)
	{
	report_error(FATAL, &Bad_infile, (int *)NULL, (char *)NULL,
							cmd_file, "TAKE" );
	--tlevel;
	}
    else
	tlinenum[tlevel] = 1;
    myfree(cmd_file);
    cmd_file = (char *)NULL;
    }
else
    tlevel = -1;
/*
 *  set up for error or SIGINT trapping
 */
#ifdef USE_SIGNAL
signal( SIGINT, sigint_handler );
#endif
if (setjmp(menu_env) != 0)
    {
    fputs("\nRETURNING TO PC-KIMMO COMMAND PROCESSING\n", stderr);
    }
/*
 *  acquire and process commands until the user wants to quit
 */
while ( do_cmd() )
    ;
}
