/*	USERFUNC.C - functions to support user interface for PC-KIMMO
 ***************************************************************************
 *
 *	void start_timing()
 *
 *	void stop_timing()
 *
 *	void show_timing()
 *
 *	int sigint_handler(sig)		(#ifdef USE_SIGNAL)
 *	int sig;
 *
 *	void cgetline(inLine, maxLen)
 *	unsigned char *inLine;
 *	int maxLen;
 *
 *	void free_result(resp)
 *	register RESULT *resp;
 *
 *	void printResults(resNodeP,outfp)
 *	RESULT *resNodeP;
 *	FILE *outfp;
 *
 *	char *set_filename(filename, default_name, extension)
 *	char *filename;
 *	char *default_name;
 *	char *extension;
 *
 *	int wanthelp(arg,msg)
 *	register char *arg;
 *	char *msg;
 *
 *	void compare_pairs(cmpfp,filename)
 *	FILE *cmpfp;
 *	char *filename;
 *
 *	void list_pairs()
 *
 *	void show_rules(num)
 *	int num;
 *
 *	void list_rules()
 *
 *	void set_rules(num,val)
 *	char *num;
 *	int val;
 *
 *	void show_status()
 *
 *	void do_system(cmd)
 *	unsigned char *cmd;
 *
 *	void do_close()
 *
 *	void do_generate(form)
 *	char *form;
 *
 *	void do_edit(filename)
 *	char *filename;
 *
 ***************************************************************************
 *	EDIT HISTORY
 *	17-Jun-88	written by Dave Smith
 *	18-Sep-89	SRMc - replace free() with myfree()
 *			     - simplify dosCommand() for portability
 *	19-Sep-89	SRMc - revised definition of getline()
 *			     - eliminated scanopt()
 *			     - revised chkRList() to allow flexible spacing
 *				in file input
 *	20-Sep-89	SRMc - protect isspace() with isascii()
 *	21-Sep-89	SRMc - change char to unsigned char for 8-bit safety
 *	23-Sep-89	SRMc - change variable names to all lowercase
 *			     - rewrote checkres(), renamed to check_result()
 *			     - rewrote putRes(), renamed to put_res()
 *			     - rewrote chkRList(), renamed to check_this()
 *	25-Sep-89	SRMc - moved pSubMenu(), cgetline(), and cmdTrace()
 *				from sh_main.c
 *			     - moved printResults() from kimmo.c
 *			     - finetune screen displays
 *	26-Sep-89	SRMc - add no_result to check_result()
 *			     - change dosCommand() to fire up the shell if
 *				the command is empty
 *	28-Sep-89	SRMc - make Lang global, remove "lang" arguments
 *			     - simplify tracing and logging variables
 *	 2-Oct-89	SRMc - rename Language to LANGUAGE
 *			     - rename ResNode to RESULT
 *	 3-Oct-89	SRMc - change interface to printResults()
 *	 4-Oct-89	SRMc - remove "Results:" from output of printResults()
 *	 6-Oct-89	SRMc - move functions from SH_UTIL.C: cgetline()
 *				check_result(), free_result(), getStr(),
 *				valForm(), and printResults()
 *	----------------------------------------------------------------------
 *	 3-Oct-89	SRMc - import code from a variety of places in
 *				previous user interface files for PC-KIMMO
 *	 4-Oct-89	SRMc - add comment character to show_status() display
 *			     - rename disp_pairs() to list_pairs()
 *			     - rename disp_rules() to list_rules()
 *			     - rename dump_rule() to show_rules()
 *	 6-Oct-89	SRMc - fix bug in set_rules() -- it wasn't computing
 *				new feasible pairs when "all" rules are set
 *			     - move handler() from usercmd.c, rename it to
 *				sigint_handler()
 *	10-Oct-89	SRMc - use report_error() for error messages
 *			     - write show_lexicon()
 *	11-Oct-89	SRMc - add Timing_flag and SET TIMING command
 *	12-Oct-89	SRMc - rename cmp_lex_surf() to compare_pairs()
 *			     - write compare_recog()
 *			     - write compare_gener()
 *			     - merged check_result() into read_results(),
 *				compare_recog() and compare_gener()
 *	13-Oct-89	SRMc - revise messages for verbose output by
 *				compare_pairs()
 *	14-Oct-89	SRMc - eliminated unused function getStr()
 *			     - moved valForm() to VALFORM.C
 *			     - imported strlower() from STRLOWER.C
 *	16-Oct-89	SRMc - finetune output from show_lex_item()
 *			     - edit error messages
 *	20-Oct-89	SRMc - simplify error messages
 *	21-Oct-89	SRMc - new interface to find_lexicon()
 *			     - new interface to generator()
 *			     - new interface to recognizer()
 *			     - new interface to getline()
 *	24-Oct-89	SRMc - show Lang.boundary instead of "***NONE***"
 *				for empty continuation in lexicon display
 *			     - some delinting
 *	30-Oct-89	SRMc - move free_result() from USERFUNC.C to PCKFUNCS.C
 *	13-Dec-89	SRMc - add filename to report_error() argument list
 *			     - don't show specificity in show_rule() display
 *			     - add do_edit()
 *	18-Dec-89	SRMc - edit error message 400, revise its handling
 *	 2-Jan-90	SRMc - use free_result() in compare_pairs()
 *			     - more delinting, add function prototypes
 *	 3-Jan-90	SRMc - will we never run out of lint?
 *	24-Jan-90	SRMc - edit error messages
 *	25-Jan-90	SRMc - edit the help messages
 *			     - port to A/UX
 *	26-Jan-90	SRMc - port to 4.2BSD (SunOS 3.2)
 *	 6-Feb-90	SRMc - add message when spawning a subshell
 *	 8-Mar-90	EA   - ifdef for THINK_C
 *	 7-May-90	JT   - added CD command for Mac version
 *	 7-May-90	JT   - fixed set timing command for Mac version
 *	12-Jul-90	SRMc - change "#ifdef AUX" to "#ifndef CLK_TCK" to
 *				protect "#define CLK_TCK", as suggested by
 *				by Greg Lee (lee@uhccux.uhcc.hawaii.edu) for
 *				port to AT&T UNIX PC 7300
 *			     - replace "void *" with "VOIDP", as also suggested
 *				by Greg Lee for port to ULTRIX
 *	26-Feb-91	SRMc - change definitions of ALTERNATION and struct
 *				lex_item to store array indices rather than
 *				pointers into arrays (that may be shifted
 *				around in memory during loading)
 *	19-Nov-91	SRMc - add fflush(stderr) for THINK C following each
 *				putc('.',stderr)
 *			     - add #ifndef MSC for errno declaration
 *	 5-Dec-91	SRMc - add prototypes for Think C 5.0
 ***************************************************************************
 * Copyright 1989, 1990, 1991 by the Summer Institute of Linguistics, Inc.
 * All rights reserved.
 */
#include <stdio.h>
#include <ctype.h>
#include <setjmp.h>
#include <errno.h>
#ifndef MSC
extern int errno;
#endif
#include "pckimmo.h"
#include "userfunc.h"

#ifdef THINK_C
#include <string.h>
#include <pascal.h>
static long start, finish;
#endif

#ifdef MSC
#include <string.h>
#include <memory.h>		/* memset() and memcpy() used in cgetline() */
#include <dos.h>		/* needed for intdosx() in cgetline() */
#include <sys/types.h>		/* for MSDOS timing functions */
#include <sys/timeb.h>
struct timeb start, finish;
#endif

#ifdef DJGPP
#include <sys/time.h>
static struct timeval t_start, t_finish;
#endif

#ifdef USG
#include <string.h>
#include <sys/types.h>		/* for UNIX System V timing functions */
#include <sys/times.h>
#include <time.h>
#ifdef __STDC__
extern long times(struct tms *buffer);
#else
extern long times();
#endif
static long startup, finish;
static struct tms tms_start, tms_finish;
#endif

#ifdef BSD
#include <strings.h>
#include <sys/time.h>
#include <sys/resource.h>
static struct timeval t_start, t_finish;
static struct rusage r_start, r_finish;
#endif

#ifdef USE_SIGNAL
#include <signal.h>
#endif

#ifdef __STDC__		/* function prototypes */

#ifndef DJGPP
extern int atoi(char *string);
extern char *getenv(char *string);
#endif

#else				/* old-fashioned function declarations */

#ifndef THINK_C
extern char *getenv();
#endif
extern int atoi();

#endif

extern LANGUAGE Lang;		/* global data for the current language */
extern unsigned char whiteSpc[];
extern int Verbose_flag;
extern int Trace_flag;
extern int Timing_flag;
extern int Limit_flag;
extern unsigned char commentChr;
extern FILE *Log_fp;
extern char *rule_file, *lex_file, *log_file;
extern jmp_buf menu_env;    /* environment to restore for error recovery */
/*
 *  error messages
 */
extern struct message Miss_argument, Bad_argument, Bad_outfile;
static struct message No_log_file =
    { 114, "No log file was open" };
static struct message Already_log_file =
    { 115, "Closing the existing log file %s" };
static struct message Missing_edit_file =
    { 116, "Missing filename for EDIT command" };
#ifdef THINK_C
static struct message Missing_directory =
    { 117, "Missing pathname for CD command" };
static struct message Directory_not_found =
    { 118, "Bad pathname for CD command" };
#endif
static struct message Input_too_long =
    { 100,  "Input line too long -- ignoring after first %d characters" };

static struct message No_comp_surf =
    { 400, "Surface form not found in pairs comparison file"};

#ifdef UNIX
#define DIRSEPCHAR '/'
#endif
#ifdef MSDOS
#define DIRSEPCHAR '\\'
#endif
#ifdef THINK_C
#define DIRSEPCHAR ':'
#endif

/*****************************************************************************
 * NAME
 *    start_timing
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Set the startup times for timing a computation.
 * RETURN VALUE
 *    none
 */
void start_timing()
{
#ifdef MSC
ftime( &start );
#endif

#ifdef DJGPP
struct timezone timezone;
gettimeofday(&t_start, &timezone);
#endif

#ifdef USG
startup = times( &tms_start );
#endif

#ifdef BSD
struct timezone timezone;
gettimeofday(&t_start, &timezone);
getrusage(RUSAGE_SELF, &r_start);
#endif

#ifdef THINK_C
start = TickCount();
#endif
}

/*****************************************************************************
 * NAME
 *    stop_timing
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Set the stopping times for timing a computation.
 * RETURN VALUE
 *    none
 */
void stop_timing()
{
#ifdef MSC
ftime( &finish );
#endif

#ifdef DJGPP
struct timezone timezone;
gettimeofday(&t_finish, &timezone);
#endif

#ifdef USG
finish = times( &tms_finish );
#endif

#ifdef BSD
struct timezone timezone;
gettimeofday(&t_finish, &timezone);
getrusage(RUSAGE_SELF, &r_finish);
#endif

#ifdef THINK_C
finish = TickCount();
#endif
}

/*****************************************************************************
 * NAME
 *    show_timing
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Display the elapsed time for a computation.
 * RETURN VALUE
 *    none
 */
void show_timing()
{
#ifdef MSC
long elapse_sec;
int elapse_frac;

elapse_frac = finish.millitm - start.millitm;
if (elapse_frac < 0)
    {
    elapse_frac += 1000;
    ++start.time;
    }
elapse_frac = (elapse_frac + 5) / 10;
elapse_sec  = finish.time - start.time;
fprintf(stderr, "  Elapsed time = %ld.%02d seconds\n",
					elapse_sec, elapse_frac );
#endif

#ifdef DJGPP
long temp, sec, frac;
/*
 * print the elapsed time, converting microseconds to rounded off hundredths
 */
temp = t_finish.tv_usec - t_start.tv_usec;	/* double precision subtract */
if (temp < 0L)
    temp += 1000000L, --sec;
sec  = t_finish.tv_sec - t_start.tv_sec;
frac = temp / 10000L;				/* round off to hundredths */
temp %= 10000L;
if (temp >= 5000L)
    ++frac;
if (frac > 100)
    frac -= 100, ++sec;
fprintf(stderr, "  Elapsed time = %ld.%02d seconds\n", sec, frac);
#endif

#ifdef USG
#ifndef CLK_TCK
/* clock always ticks in 60ths of a second according to times(2) for A/UX */
#define CLK_TCK 60
#endif
long elapse, sec, frac;

if ((startup != -1L) && (finish != -1L))
    {
    elapse = finish - startup;
    sec = elapse / CLK_TCK;
    frac = (100 * (elapse % CLK_TCK)) / CLK_TCK;
    fprintf(stderr, "  Elapsed time = %ld.%02ld seconds", sec, frac);

    elapse = tms_finish.tms_utime - tms_start.tms_utime;
    sec = elapse / CLK_TCK;
    frac = (100 * (elapse % CLK_TCK)) / CLK_TCK;
    fprintf(stderr, "  (CPU time = %ld.%02ldu", sec, frac);

    elapse = tms_finish.tms_stime - tms_start.tms_stime;
    sec = elapse / CLK_TCK;
    frac = (100 * (elapse % CLK_TCK)) / CLK_TCK;
    fprintf(stderr, " + %ld.%02lds)\n", sec, frac);
    }
#endif

#ifdef THINK_C
/* clock always ticks in 60ths of a second */
#define CLK_TCK 60
long elapse, sec, frac;

elapse = finish - start;
sec = elapse / CLK_TCK;
frac = (100 * (elapse % CLK_TCK)) / CLK_TCK;
fprintf(stderr, "  Elapsed time = %ld.%02ld seconds\n", sec, frac);
#endif

#ifdef BSD
long temp, sec, frac;
/*
 * print the elapsed time, converting microseconds to rounded off hundredths
 */
temp = t_finish.tv_usec - t_start.tv_usec;	/* double precision subtract */
if (temp < 0L)
    temp += 1000000L, --sec;
sec  = t_finish.tv_sec - t_start.tv_sec;
frac = temp / 10000L;				/* round off to hundredths */
temp %= 10000L;
if (temp >= 5000L)
    ++frac;
if (frac > 100)
    frac -= 100, ++sec;
fprintf(stderr, "  Elapsed time = %ld.%02d seconds", sec, frac);
/*
 *  follow the same algorithm for printing the user and system times
 */
temp = r_finish.ru_utime.tv_usec - r_start.ru_utime.tv_usec;
if (temp < 0L)					/* double precision subtract */
    temp += 1000000L, --sec;
sec  = r_finish.ru_utime.tv_sec - r_start.ru_utime.tv_sec;
frac = temp / 10000L;				/* round off to hundredths */
temp %= 10000L;
if (temp > 5000L)
    ++frac;
if (frac > 100)
    frac -= 100, ++sec;
fprintf(stderr, "  (CPU time = %ld.%02ldu", sec, frac);

temp = r_finish.ru_stime.tv_usec - r_start.ru_stime.tv_usec;
if (temp < 0L)					/* double precision subtract */
    temp += 1000000L, --sec;
sec  = r_finish.ru_stime.tv_sec - r_start.ru_stime.tv_sec;
frac = temp / 10000L;				/* round off to hundredths */
temp %= 10000L;
if (temp > 5000L)
    ++frac;
if (frac > 100)
    frac -= 100, ++sec;
fprintf(stderr, " + %ld.%02lds)\n", sec, frac);
#endif
}

#ifdef USE_SIGNAL
/****************************************************************************
 * NAME
 *    sigint_handler
 * ARGUMENTS
 *    sig - should always be SIGINT
 * DESCRIPTION
 *    Our handler for SIGINT signals (control-C's from keyboard).
 * RETURN VALUE
 *    none
 */
/*ARGSUSED*/
#ifdef __STDC__
void
#else
int
#endif
 sigint_handler(sig)
int sig;
{
extern FILE *Fileout_fp;
    
if (Fileout_fp != (FILE *)NULL)
    {
    fflush(Fileout_fp);
    fclose(Fileout_fp);
    }
signal(SIGINT,sigint_handler);		/* restore SIGINT handling */
longjmp(menu_env,1);			/* go back to the main menu */
}
#endif

/****************************************************************************
 * NAME
 *    cgetline
 * ARGUMENTS
 *    inLine - pointer to input buffer
 *    maxLen - maximum number of characters to store in inLine (not counting
 *		the terminating NUL)
 * DESCRIPTION
 *    read a line of input from the keyboard
 * RETURN VALUE
 *    none
 */
void cgetline(inLine, maxLen)
unsigned char *inLine;
int  maxLen;
{
#ifdef MSC
char far *pbuf;
unsigned char buffer[MAXLINELEN];
union REGS inreg;
union REGS outreg;
struct SREGS segreg;
int num;

signal( SIGINT, SIG_IGN );	/* ignore SIGINT during keyboard input */
memset(&inreg,  NUL, sizeof(union REGS));	/* clear memory */
memset(&outreg, NUL, sizeof(union REGS));	/* (for paranoia) */
memset(&segreg, NUL, sizeof(struct SREGS));
memset(buffer,  NUL, MAXLINELEN);
pbuf = buffer;				/* prepare for DOS call */
inreg.h.ah = 0x0A;			/* int 21H, func 0AH */
inreg.x.dx = FP_OFF(pbuf);
segreg.ds = FP_SEG(pbuf);
buffer[0] = MAXLINELEN - 3;

intdosx(&inreg, &outreg, &segreg);	/* call DOS for an input line */
					/* this allows fancy command line */
					/* editors like CED and NDOSEDIT */
					/* to be used inside PCKIMMO */

putc('\n',stderr);			/* provide an extra newline */
num = buffer[1];			/* copy to the caller's buffer */
if (num > maxLen)
    num = maxLen;
memcpy(inLine, buffer+2, num);
inLine[num] = NUL;			/* make sure NUL-terminated */
signal( SIGINT, sigint_handler );	/* restore SIGINT trapping */

#else /*MSC*/

register int c;

#ifdef USE_SIGNAL
signal( SIGINT, SIG_IGN );	/* ignore SIGINT during keyboard input */
#endif
fgets((char *)inLine, maxLen, stdin);
inLine[maxLen] = NUL;
if (strchr((char *)inLine,'\n') == (char *)NULL)
    {
    /*
     *  flush the rest of the input line after reading the maximum number
     *    of characters
     */
    while (((c = getc(stdin)) != EOF) && (c != '\n'))
	;
    if (c != EOF)
	report_error(NONFATAL, &Input_too_long, (int *)NULL, (char *)NULL,
							maxLen );
    }
#ifdef USE_SIGNAL
signal( SIGINT, sigint_handler );	/* restore SIGINT trapping */
#endif
#endif /*else MSC*/
}

/****************************************************************************
 * NAME
 *    print_result
 * ARGUMENTS
 *    nodep   - pointer to result node
 *    str     - pointer to comment string
 *    linenr  - file line number
 *    outfp   - output FILE pointer
 * DESCRIPTION
 *    Print one line of the comparison result to the specified file.
 * RETURN VALUE
 *    none
 */
static void print_result( nodep, str, linenr, outfp)
RESULT *nodep;
char *str;
int linenr;
FILE *outfp;
{
register int k;

if (nodep == (RESULT *)NULL)
    {
    if (str == (char *)NULL)
	{			/* blank line with only the line number */
	fprintf(outfp, "\n%3d.", linenr );
	}
    else			/* just the comment string (form?) */
	fprintf(outfp, "\n%3d. %s", linenr, str );
    }
else
    {
    fprintf( outfp, "\n%3d.", linenr);
    k = 0;
    if (nodep->str && *nodep->str)
	{
	fprintf( outfp, "   %s", nodep->str);
	k += strlen((char *)nodep->str) + 3;
	}
    if (nodep->feat && *nodep->feat)
	{
	fprintf( outfp, "   %s", nodep->feat);
	k += strlen((char *)nodep->feat) + 3;
	}
    if (str && *str)
	{
	while (k++ < 40)
	    putc( ' ', outfp );
	fprintf( outfp, " %s", str);
	}
    }
}

/****************************************************************************
 * NAME
 *    put_res
 * ARGUMENTS
 *    nodep   - pointer to result node
 *    str     - pointer to comment string
 *    linenr  - file line number
 * DESCRIPTION
 *    Print one line of the comparison result, to both the screen and the
 *    log file (if any).
 * RETURN VALUE
 *    none
 */
static void put_res( nodep, str, linenr)
RESULT *nodep;
char *str;
int linenr;
{
if (Log_fp != (FILE *)NULL)
    print_result( nodep, str, linenr, Log_fp );
print_result( nodep, str, linenr, stderr );
}

/****************************************************************************
 * NAME
 *    check_this
 * ARGUMENTS
 *    thisp - pointer to current result
 *    listp - pointer to list of computed results
 * DESCRIPTION
 *    Check the result pointed to by thisp against the list of results
 *    pointed to by listp.
 * RETURN VALUE
 *    none
 */
static void check_this( thisp, listp )
RESULT *thisp;
RESULT *listp;
{
RESULT *rnp;
unsigned char *this_str, *this_feat;
unsigned char *res_str, *res_feat;

if (thisp == (RESULT *)NULL)
    return;
this_str  = (thisp->str != (unsigned char *)NULL) ?
			thisp->str  : (unsigned char *)"";
this_feat = (thisp->feat != (unsigned char *)NULL) ?
			thisp->feat : (unsigned char *)"";

for ( rnp = listp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
    {
    res_str  = (rnp->str != (unsigned char *)NULL) ?
			rnp->str  : (unsigned char *)"";
    res_feat = (rnp->feat != (unsigned char *)NULL) ?
			rnp->feat : (unsigned char *)"";
    /*
     *  compare result strings
     */
    if (strcmp((char *)this_str,(char *)res_str) != 0)
	continue;			/* strings don't match */
    /*
     *  compare features
     */
    if (strcmp((char *)this_feat,(char *)res_feat) != 0)
	continue;			/* features don't match */
    /*
     *  at this point, we have a good match
     */
    thisp->okay = TRUE;			/* we have a match, so mark these */
    rnp->okay = TRUE;			/*   results as valid */
    return;				/* no need to look further */
    }
}

/****************************************************************************
 * NAME
 *    read_results
 * ARGUMENTS
 *    cmpfp    - input FILE pointer for comparison file
 *    lnump    - pointer to input file line number counter
 * DESCRIPTION
 *    Read a list of results from a file.
 * RETURN VALUE
 *    pointer to a linked list of RESULTs, or NULL if none
 */
static RESULT *read_results( cmpfp, lnump)
FILE *cmpfp;
int *lnump;
{
unsigned char *buf;
unsigned char *p;
unsigned char *pe;
RESULT *rnp;
RESULT *filresp;	/* list of results read from the file */
RESULT *lastp;
/*
 *  read the result line from the comparison file, one line at a time
 */
filresp = (RESULT *)NULL;
lastp   = (RESULT *)NULL;
while ((buf = getline(cmpfp, lnump, commentChr)) != (unsigned char *)NULL)
    {
    /*
     *  1. move past leading spaces
     *  2. truncate trailing spaces
     *  3. quit if blank line (marks end of this word's data)
     */
    for ( p = buf ; *p && isascii(*p) && isspace(*p) ; ++p )
	;
    for (   pe = p + strlen((char *)p) - 1 ;
	    (pe >= p) && isascii(*pe) && isspace(*pe) ;
	    *pe-- = NUL )
	;
    if ( *p == NUL )
	break;
    /*
     *  add this result to the list of results from the file
     */
    rnp = (RESULT *)myalloc(sizeof(RESULT));
    for ( pe = p ; *pe && !(isascii(*pe) && isspace(*pe)) ; ++pe )
	;
    if (*pe == NUL)
	{					 /* save the result string */
	rnp->str = (unsigned char *)mystrdup((char *)p);
	rnp->feat = (unsigned char *)NULL;	/* no feature string */
	}
    else
	{
	*pe++ = NUL;				/* save the result string */
	rnp->str = (unsigned char *)mystrdup((char *)p);
	for (p=pe ; *p && isascii(*p) && isspace(*p) ; ++p)
	    ;				/* skip whitespace until feature */
					/* save the feature string */	
	rnp->feat = (unsigned char *)mystrdup((char *)p);
	}
    if (filresp == (RESULT *)NULL)
	filresp = rnp;			/* first node in the list */
    else
	lastp->link = rnp;		/* add to the end of the list */
    lastp = rnp;
    lastp->okay = FALSE;
    lastp->link = (RESULT *)NULL;
    }
return( filresp );
}

/****************************************************************************
 * NAME
 *    printResults
 * ARGUMENTS
 *    resp  - pointer to linked list of results
 *    outfp - output FILE pointer
 * DESCRIPTION
 *    Write the results to the indicated output file.
 * RETURN VALUE
 *    none
 */
void printResults(resp,outfp)
RESULT *resp;
FILE *outfp;
{
if (resp == (RESULT *)NULL)
    fprintf(outfp, "  *** NONE ***\n");
else
    {
    while ( resp != (RESULT *)NULL )
	{
	fprintf(outfp, "  %s", resp->str);
	if (resp->feat != (unsigned char *)NULL)
	    fprintf(outfp, "     %s", resp->feat);
	putc('\n', outfp);
	resp = resp->link;
	}
    }
putc('\n', outfp);
}

/*****************************************************************************
 * NAME
 *    set_filename
 * ARGUMENTS
 *    filename  - filename given by user (may be NULL)
 *    default_name   - default full filename (may be NULL)
 *    extension - default extension, including leading separation char (.)
 * DESCRIPTION
 *    Check a filename provided by the user, appending a default extension
 *    if the file lacks one, or even providing a default filename.
 * RETURN VALUE
 *    pointer to a dynamically allocated filename string
 */
char *set_filename(filename, default_name, extension)
char *filename;
char *default_name;
char *extension;
{
register char *p;

if (filename == (char *)NULL)
    {
    if (default_name == (char *)NULL)
	return( (char *)NULL );
    else
	return( (char *)mystrdup(default_name) );
    }
#ifdef DIRSEPCHAR
/*
 *  find the beginning of the filename proper, following an optional
 *    string of directory names separated by DIRSEPCHAR
 */
p = (char *)strrchr(filename, DIRSEPCHAR);
if (p != (char *)NULL)
    ++p;
else
#endif
    p = filename;

if (strchr(p, *extension) == (char *)NULL)
    {
    return(strcat(strcpy((char *)myalloc((unsigned)
				      (strlen(filename)+strlen(extension)+1)),
			 filename),
		  extension) );
    }
return( (char *)mystrdup(filename) );
}

/*****************************************************************************
 * NAME
 *    wanthelp
 * ARGUMENTS
 *    arg - pointer to argument string
 *    msg - pointer to message string
 * DESCRIPTION
 *    If the argument string is "?", then print the helpful message.
 * RETURN VALUE
 *    1 if help wanted (and message printed), 0 otherwise
 */
int wanthelp(arg,msg)
register char *arg;
char *msg;
{
if (arg == NULL)
    return(0);
if ((*arg++ == '?') && (*arg == NUL))
    {
    fprintf(stderr, "%s\n", msg);
    return(1);
    }
return(0);
}

/****************************************************************************
 * NAME
 *    compare_gener
 * ARGUMENTS
 *    cmpfp - input comparison FILE pointer
 * DESCRIPTION
 *    Execute a FILE COMPARE GENERATE <filename> command.
 * RETURN VALUE
 *    none
 */
void compare_gener(cmpfp)
FILE *cmpfp;
{
int linenum, formline, mismatch, num_dots;
unsigned char *p, *form;
RESULT *resp, *filresp, *rnp;
RESULT no_result;

linenum = 1;
num_dots = 78;		/* print 78 dots before a newline */
if (Timing_flag)
    start_timing();
while ((p = getline(cmpfp, &linenum, commentChr)) != (unsigned char *)NULL)
    {
    if (*p != NUL)
	{
	/*
	 *  compute the results for the form read from the file
	 *  set all the computed results to "not okay" to begin with
	 */
	form = (unsigned char *)mystrdup( (char *)strtok8(p, whiteSpc) );
	formline = linenum - 1;
	resp = generator(form, &Lang,Limit_flag,Trace_flag,Log_fp);
	for ( rnp = resp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
	    rnp->okay = FALSE;
	/*
	 *  read the prepared results for comparison
	 *  for each result from the file, scan the computed result list for a
	 *    match
	 */
	filresp = read_results(cmpfp, &linenum);
	for ( rnp = filresp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
	    check_this( rnp, resp );
	/*
	 *  check for errors in the analysis
	 */
	mismatch = 0;
	for ( rnp = filresp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
	    {
	    if (rnp->okay == FALSE)
		++mismatch;
	    }
	for ( rnp = resp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
	    {
	    if (rnp->okay == FALSE)
		++mismatch;
	    }
	/*
	 *  list the results, verbosely if needed
	 */
	if (Verbose_flag || mismatch || (resp == (RESULT *)NULL))
	    {
	    num_dots = 0;	/* force newline later if not verbose */
	    /*
	     *  print the form read from the file
	     */
	    put_res( (RESULT *)NULL, (char *)form, formline++ );
	    /*
	     *  print the results from the file
	     */
	    for ( rnp = filresp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
		{
		if (rnp->okay == TRUE)
		    put_res( rnp, ":OK", formline++ );
		else
		    put_res( rnp, ":Not in result list", formline++ );
		}
	    /*
	     *  print any computed results which weren't in the file
	     */
	    for ( rnp = resp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
		{
		if (rnp->okay == FALSE)
		    put_res( rnp, ":In result list but not in file", formline);
		}
	    /*
	     *  if no results, say so.
	     */
	    if (resp == (RESULT *)NULL)
		{
		no_result.str  = (unsigned char *)"*** NONE ***";
		no_result.feat = (unsigned char *)"";
		put_res( &no_result, (char *)NULL, formline);
		}
	    /*
	     *  print a blank line (only the line number showing)
	     */
	    put_res( (RESULT *)NULL, (char *)NULL, formline);
	    }
	else
	    {
	    if (num_dots-- > 0)
		{
		putc('.', stderr);
#ifdef THINK_C
		fflush(stderr);
#endif
		}
	    else
		{
		fprintf(stderr, "\n.");
		num_dots = 77;		/* one dot of 78 already printed */
		}
	    }
	/*
	 *  release memory used for this comparison
	 */
	free_result( resp );
	free_result( filresp );
	myfree((char *)form);
	}
    }
if (Timing_flag)
    stop_timing();
putc('\n',stderr);
}

/****************************************************************************
 * NAME
 *    compare_recog
 * ARGUMENTS
 *    cmpfp - input comparison FILE pointer
 * DESCRIPTION
 *    Execute a FILE COMPARE RECOGNIZE <filename> command.
 * RETURN VALUE
 *    none
 */
void compare_recog(cmpfp)
FILE *cmpfp;
{
int linenum, formline, mismatch, num_dots;
unsigned char *p, *form;
RESULT *resp, *filresp, *rnp;
RESULT no_result;

linenum = 1;
num_dots = 78;		/* print 78 dots before a newline */
if (Timing_flag)
    start_timing();
while ((p = getline(cmpfp, &linenum, commentChr)) != (unsigned char *)NULL)
    {
    if (*p != NUL)
	{
	/*
	 *  compute the results for the form read from the file
	 *  set all the computed results to "not okay" to begin with
	 */
	form = (unsigned char *)mystrdup( (char *)strtok8(p, whiteSpc) );
	formline = linenum - 1;
	resp = recognizer(form, &Lang,Limit_flag,Trace_flag,Log_fp);
	for ( rnp = resp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
	    rnp->okay = FALSE;
	/*
	 *  read the prepared results for comparison
	 *  for each result from the file, scan the computed result list for a
	 *    match
	 */
	filresp = read_results(cmpfp, &linenum);
	for ( rnp = filresp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
	    {
	    rnp->okay = FALSE;
	    check_this( rnp, resp );
	    }
	/*
	 *  check for errors in the analysis
	 */
	mismatch = 0;
	for ( rnp = filresp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
	    {
	    if (rnp->okay == FALSE)
		++mismatch;
	    }
	for ( rnp = resp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
	    {
	    if (rnp->okay == FALSE)
		++mismatch;
	    }
	/*
	 *  list the results, verbosely if needed
	 */
	if (Verbose_flag || mismatch || (resp == (RESULT *)NULL))
	    {
	    num_dots = 0;	/* force newline later if not verbose */
	    /*
	     *  print the form read from the file
	     */
	    put_res( (RESULT *)NULL, (char *)form, formline++ );
	    /*
	     *  print the results from the file
	     */
	    for ( rnp = filresp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
		{
		if (rnp->okay == TRUE)
		    put_res( rnp, ":OK", formline++ );
		else
		    put_res( rnp, ":Not in result list", formline++ );
		}
	    /*
	     *  print any computed results which weren't in the file
	     */
	    for ( rnp = resp ; rnp != (RESULT *)NULL ; rnp = rnp->link )
		{
		if (rnp->okay == FALSE)
		    put_res( rnp, ":In result list but not in file", formline);
		}
	    /*
	     *  if no results, say so.
	     */
	    if (resp == (RESULT *)NULL)
		{
		no_result.str  = (unsigned char *)"*** NONE ***";
		no_result.feat = (unsigned char *)"";
		put_res( &no_result, (char *)NULL, formline);
		}
	    /*
	     *  print a blank line (only the line number showing)
	     */
	    put_res( (RESULT *)NULL, (char *)NULL, formline);
	    }
	else
	    {
	    if (num_dots-- > 0)
		{
		putc('.', stderr);
#ifdef THINK_C
		fflush(stderr);
#endif
		}
	    else
		{
		fprintf(stderr, "\n.");
		num_dots = 77;		/* one dot of 78 already printed */
		}
	    }
	/*
	 *  release memory used for this comparison
	 */
	free_result( resp );
	free_result( filresp );
	myfree((char *)form);
	}
    }
if (Timing_flag)
    stop_timing();
putc('\n',stderr);
}

static int chk_res( resp, form )
RESULT *resp;
unsigned char *form;
{	
int count;

for ( count = 0 ; resp != (RESULT *)NULL ; resp = resp->link )
    {
    if (strcmp((char *)resp->str, (char *)form) == 0)
	++count;
    }
return( count );
}

/****************************************************************************
 * NAME
 *    compare_pairs
 * ARGUMENTS
 *    cmpfp    - input comparison FILE pointer
 *    filename - input comparison file name (for error messages)
 * DESCRIPTION
 *    Execute a FILE COMPARE PAIRS <filename> command.
 * RETURN VALUE
 *    none
 */
void compare_pairs(cmpfp,filename)
FILE *cmpfp;
char *filename;
{
register unsigned char *p;
int surf_cnt, surf_errs, surf_ok;
int lex_cnt, lex_errs, lex_ok;
int num;
int linenum;
int num_forms, formline;
int num_dots;
int had_error;
RESULT lex_form, surf_form, *lex_resp, *surf_resp, *rp;
char ambigbuf[MAXLINELEN];

surf_errs = 0;	surf_ok = 0;
lex_errs = 0;	lex_ok = 0;
num_forms = 0;
linenum = 1;
num_dots = 78;		/* print 78 dots before a newline */
had_error = FALSE;

if (Timing_flag)
    start_timing();
while ( (p = getline(cmpfp,&linenum, commentChr)) != (unsigned char *)NULL )
    {
    p = strtok8(p, whiteSpc);
    if (p != (unsigned char *)NULL)
	{
	lex_form.str = (unsigned char *)mystrdup((char *)p);
	lex_form.feat = (unsigned char *)NULL;
	lex_form.link = (RESULT *)NULL;
	formline = linenum - 1;
	p = getline(cmpfp,&linenum,commentChr);
	if (p == (unsigned char *)NULL)
	    {
	    --linenum;
	    report_error(NONFATAL, &No_comp_surf, &linenum, filename);
	    break;			/* exit loop since EOF */
	    }
	p = strtok8(p, whiteSpc);
	if (p == (unsigned char *)NULL)
	    {
	    --linenum;
	    report_error(NONFATAL, &No_comp_surf, &linenum, filename);
	    continue;
	    }
	surf_form.str = (unsigned char *)mystrdup((char *)p);
	surf_form.feat = (unsigned char *)NULL;
	surf_form.link = (RESULT *)NULL;
	/*
	 *  at this point, we have a lex/surf pair
	 */
	surf_resp=generator(lex_form.str,&Lang,Limit_flag,Trace_flag,Log_fp);
	lex_resp =recognizer(surf_form.str,&Lang,Limit_flag,Trace_flag,Log_fp);
	/*
	 *  keep track for summary
	 */
	++num_forms;
	surf_cnt = chk_res( surf_resp, surf_form.str );
	lex_cnt = chk_res( lex_resp,  lex_form.str );

	if (surf_cnt == 0)
	    ++surf_errs;
	else
	    ++surf_ok;
	if (lex_cnt == 0)
	    ++lex_errs;
	else
	    ++lex_ok;
	/*
	 *  report verbose results, if requested or required
	 */
	if ( Verbose_flag || (lex_cnt==0 ) || (surf_cnt==0) )
	    {
	    num_dots = 0;	/* force newline later if not verbose */
	    /*
	     *  display the lexical form read from the file
	     */
	    if (lex_cnt == 0)
		put_res( &lex_form, ":Not recognized from surface form",
						formline++ );
	    else if ((lex_cnt == 1) && (lex_resp->link == (RESULT *)NULL))
		put_res( &lex_form, ":OK", formline++ );
	    else
		{
		num = 0;
		for ( rp = lex_resp ; rp != (RESULT *)NULL ; rp = rp->link )
		    ++num;
		sprintf(ambigbuf,
			":OK (matches %d of %d recognized)",
			lex_cnt, num);
		put_res( &lex_form, ambigbuf, formline++ );
		}
	    /*
	     *  display the surface form read from the file
	     */
	    if (surf_cnt == 0)
		put_res( &surf_form, ":Not generated from lexical form",
						formline++ );
	    else if ((surf_cnt == 1) && (surf_resp->link == (RESULT *)NULL))
		put_res( &surf_form, ":OK", formline++ );
	    else
		{
		num = 0;
		for ( rp = surf_resp ; rp != (RESULT *)NULL ; rp = rp->link )
		    ++num;
		sprintf(ambigbuf,
			":OK (matches %d of %d generated)",
			surf_cnt, num);
		put_res( &surf_form, ambigbuf, formline++ );
		}
	    }
	else
	    {
	    if (num_dots-- > 0)
		{
		putc('.', stderr);
#ifdef THINK_C
		fflush(stderr);
#endif
		}
	    else
		{
		fprintf(stderr, "\n.");
		num_dots = 77;		/* one dot of 78 already printed */
		}
	    }
	had_error = !(surf_cnt && lex_cnt);
	/*
	 *  release memory used for this comparison
	 */
	free_result( surf_resp );
	free_result( lex_resp );
	}
    else
	{
	if (Verbose_flag || had_error)
	    put_res( (RESULT *)NULL, (char *)NULL, linenum-1 );
	}
    }
if (Timing_flag)
    stop_timing();

if (Log_fp != (FILE *)NULL)
    {
    fprintf(Log_fp, "\nSummary of results for %d forms:", num_forms);
    fprintf(Log_fp, "\n   Recognizer: %3d errors  %3d ok", lex_errs, lex_ok);
    fprintf(Log_fp, "\n   Generator:  %3d errors  %3d ok", surf_errs, surf_ok);
    }
fprintf(stderr, "\nSummary of results for %d forms:\n", num_forms);
fprintf(stderr, "   Recognizer: %3d errors  %3d ok\n", lex_errs, lex_ok);
fprintf(stderr, "   Generator:  %3d errors  %3d ok\n", surf_errs, surf_ok);
}

/****************************************************************************
 * NAME
 *    list_pairs
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Execute a LIST PAIRS command.
 * RETURN VALUE
 *    none
 */
#define MPL 39
void list_pairs()
{
int i, num, left;
unsigned char *ps, *pl;

if ( Lang.lex_pair == (unsigned char *)NULL )
    return;
/*
 *  display the feasible pairs
 */
ps = Lang.surf_pair;
pl = Lang.lex_pair;
left = Lang.num_pairs;
while ( left > 0 )
    {
    num = (MPL < left) ? MPL : left;
    for ( i = 0 ; i < num ; i++ )
	fprintf(stderr, "%c ", *pl++);
    putc( '\n', stderr);
    for ( i = 0 ; i < num ; i++ )
	fprintf(stderr, "%c ", *ps++);
    putc( '\n', stderr);
    left -= num;
    if (left > 0)
	putc( '\n', stderr);	/* separate each set of two lines */
    }
putc( '\n', stderr);		/* blank line before the prompt */
}

/****************************************************************************
 * NAME
 *    show_rules
 * ARGUMENTS
 *    num - number of the rule to dump
 * DESCRIPTION
 *    Execute a SHOW RULES <number> command.
 *    (Despite the function name, only one rule is shown.)
 * RETURN VALUE
 *    none
 */
void show_rules(num)
int num;
{
register int i;
RULE *rp;
struct fsa_column *cp;
unsigned char *lp, *sp;

if ((num <= 0) || (num > Lang.num_rules))
    return;			/* should be checked by caller, but ... */
rp = &Lang.automata[num-1];
fprintf(stderr, "%3d %s %s\n",
			num, (rp->rule_active) ? "on " : "off", rp->name );
if (!rp->rule_active)
    return;

for ( cp = rp->columns, i = 0 ; i < rp->num_cols ; ++i, ++cp )
    {
    fprintf(stderr, "%s:%s  (", cp->lex_name, cp->surf_name );
    for ( lp = cp->lex_chars, sp = cp->surf_chars ; *lp != NUL ; ++lp, ++sp )
	fprintf(stderr, " %c:%c", *lp, *sp);
    fprintf(stderr, " )\n" );
    }
}

/****************************************************************************
 * NAME
 *    list_rules
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Execute a LIST RULES command.
 * RETURN VALUE
 *    none
 */
void list_rules()
{
register int i;
int num;
RULE *rp;

for ( num = Lang.num_rules, rp = Lang.automata, i = 0 ; i < num ; ++i, ++rp )
    {
    fprintf(stderr, "%3d %s %s\n",
		i+1, (rp->rule_active) ? "on " : "off", rp->name );
    }
}

/*****************************************************************************
 * NAME
 *    set_rules
 * ARGUMENTS
 *    num - number string of the first rule to set, or NULL for all rules
 *    val - value for setting: TRUE for ON, FALSE for OFF
 * DESCRIPTION
 *    Execute a SET RULES {ON|OFF} {ALL|<number(s)>} command.
 * RETURN VALUE
 *    none
 */
void set_rules(num,val)
char *num;
int val;
{
register int i;

if (num == (char *)NULL)
    {
    for ( i = 0 ; i < Lang.num_rules ; ++i )
	Lang.automata[i].rule_active = val;
    }
else do
    {
    i = atoi(num);
    if ((i > 0) && (i <= Lang.num_rules))
	Lang.automata[i-1].rule_active = val;
    else
	break;
    } while (	((num = (char *)strtok8((unsigned char *)NULL,whiteSpc)) !=
							       (char *)NULL) &&
		isascii(*num) && isdigit(*num) );
if (num != (char *)NULL)
    report_error(FATAL, &Bad_argument, (int *)NULL, (char *)NULL,
							"SET RULE", num);
feasible_pairs(&Lang);
}

/*****************************************************************************
 * NAME
 *    show_status
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Execute a SHOW STATUS command.
 * RETURN VALUE
 *    none
 */
void show_status()
{
fprintf(stderr, "Status of PC-KIMMO:\n");
fprintf(stderr, "    Rules file:   %s\n",
		(rule_file != (char *)NULL) ? rule_file : "(none)" );
fprintf(stderr, "    Lexicon file: %s\n",
		(lex_file != (char *)NULL) ? lex_file : "(none)" );
fprintf(stderr, "    Log file:     %s\n",
		(log_file != (char *)NULL) ? log_file : "(none)" );
fprintf(stderr, "    Comment character is %c\n", commentChr );
fprintf(stderr, "    Limit is %s\n", (Limit_flag)?"ON":"OFF");
fprintf(stderr, "    Timing is %s\n", (Timing_flag)?"ON":"OFF");
fprintf(stderr, "    Tracing level is %d\n", Trace_flag );
fprintf(stderr, "    Verbose comparison is %s\n", (Verbose_flag)?"ON":"OFF");
putc( '\n', stderr);		/* blank line before the prompt */
}

#ifndef THINK_C
/*****************************************************************************
 * NAME
 *    do_system
 * ARGUMENTS
 *    cmd - pointer to system command string to execute
 * DESCRIPTION
 *    Execute a SYSTEM <command> command.
 * RETURN VALUE
 *    none
 */
void do_system(cmd)
unsigned char *cmd;
{
int res;
#ifdef DJGPP
unsigned char command[200];
#endif
/*
 *  the input string may begin with whitespace -- skip past it if so
 */
if (cmd != (unsigned char *)NULL)
    {
    while (isascii(*cmd) && isspace(*cmd))
	++cmd;
    }
/*
 *  if empty command, find the appropriate shell (command processor)
 */
if ((cmd == (unsigned char *)NULL) || (*cmd == NUL))
    {
#ifdef MSC
    cmd = getenv("SHELL");
    if (cmd == NULL)
	cmd = getenv("COMSPEC");
    if (cmd == NULL)
	cmd = "COMMAND";
    fprintf(stderr,"Type EXIT to return to PC-KIMMO\n");
#endif
#ifdef DJGPP
    cmd = (unsigned char *)getenv("SHELL");
    if (cmd == NULL)
	cmd = (unsigned char *)getenv("COMSPEC");
    if (cmd == NULL)
	cmd = (unsigned char *)"COMMAND";
    fprintf(stderr,"Type EXIT to return to PC-KIMMO\n");
#endif
#ifdef UNIX
    cmd = (unsigned char *)getenv("SHELL");
    if (cmd == (unsigned char *)NULL)
	cmd = (unsigned char *)"/bin/sh";
    fprintf(stderr,"Type exit to return to PC-KIMMO\n");
#endif
    }
#ifdef DJGPP
else
    {
    strcpy((char *)command,"command /c ");
    strcat((char *)command, (char *)cmd);
    cmd = command;
    }
#endif
/*
 *  try to execute the command
 */
res = system((char *)cmd);
if ( res < 0 )
    perror("PCKIMMO");
}
#endif  /*THINK_C*/

/*****************************************************************************
 * NAME
 *    do_close
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Execute a CLOSE command.
 * RETURN VALUE
 *    none
 */
void do_close()
{
if ((Log_fp == (FILE *)NULL) && (log_file == (char *)NULL))
    {
    report_error(FATAL, &No_log_file, (int *)NULL, (char *)NULL);
    return;
    }
if (Log_fp != (FILE *)NULL)
    {
    fclose(Log_fp);
    Log_fp = (FILE *)NULL;
    }
if (log_file != (char *)NULL)
    {
    myfree(log_file);
    log_file = (char *)NULL;
    }
}

/*****************************************************************************
 * NAME
 *    do_generate
 * ARGUMENTS
 *    form - pointer to lexical form
 * DESCRIPTION
 *    Execute a GENERATE [<form>] command.
 *    If a form is given, generate the corresponding surface form.  Otherwise,
 *    prompt for lexical forms, generating the surface forms, until the user
 *    types a blank line.
 * RETURN VALUE
 *    none
 */
void do_generate(form)
unsigned char *form;
{
register unsigned char *bufp;
register RESULT *resp;
unsigned char buf[MAXLINELEN+1];

if (form != (unsigned char *)NULL)
    {
    /*
     *  form given with command -- process only that form
     */
    bufp = form;
    while (isascii(*bufp) && isspace(*bufp))
	bufp++;
    if (Log_fp != (FILE *)NULL)
	fprintf(Log_fp, "%s\n", bufp);

    if (Timing_flag)
	start_timing();

    resp = generator(bufp, &Lang,Limit_flag,Trace_flag,Log_fp);

    if (Timing_flag)
	stop_timing();

    printResults( resp, stderr );
    if (Log_fp != (FILE *)NULL)
	printResults( resp, Log_fp );
    free_result(resp);

    if (Timing_flag)
	show_timing();

    return;
    }
for (;;)
    {
    /*
     *  no form given with command -- read and process forms until blank line
     */
    fputs( "generator>>", stderr);
    cgetline(buf,MAXLINELEN);
    bufp = strtok8(buf, whiteSpc);
    if (bufp == (unsigned char *)NULL)
	return;			/* quit at first blank line */
    if (Log_fp != (FILE *)NULL)
	fprintf(Log_fp, "%s\n", bufp);

    if (Timing_flag)
	start_timing();

    resp = generator(bufp, &Lang,Limit_flag,Trace_flag,Log_fp);

    if (Timing_flag)
	stop_timing();

    printResults( resp, stderr );
    if (Log_fp != (FILE *)NULL)
	printResults( resp, Log_fp );
    free_result(resp);

    if (Timing_flag)
	show_timing(); 
    }
}

/*****************************************************************************
 * NAME
 *    do_log
 * ARGUMENTS
 *    filename - pointer to an optional log file name.
 * DESCRIPTION
 *    Execute a LOG [<filename>] command.
 * RETURN VALUE
 *    none
 */
void do_log(filename)
char *filename;
{
if (wanthelp(filename,"LOG [<file>] (default is pckimmo.log)"))
    return;
if (Log_fp != (FILE *)NULL)
    fclose(Log_fp);
if (log_file != (char *)NULL)
    {
    report_error(NONFATAL, &Already_log_file, (int *)NULL, (char *)NULL,
							log_file);
    myfree(log_file);
    log_file = (char *)NULL;
    }
if (filename == (char *)NULL)
    filename = "pckimmo.log";		/* default log file name */
Log_fp = fopen(filename,"w");
if (Log_fp == (FILE *)NULL)
    report_error(FATAL, &Bad_outfile, (int *)NULL, (char *)NULL,
							filename, "LOG" );
else
    log_file = (char *)mystrdup(filename);
}

/*****************************************************************************
 * NAME
 *    do_new
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Execute a NEW command.
 * RETURN VALUE
 *    none
 */
void do_new()
{
free_rules(&Lang);			/* remove existing rules */
if (rule_file != (char *)NULL)
    {
    myfree(rule_file);
    rule_file = (char *)NULL;
    }
free_lexicons(&Lang);			/* remove existing lexicon */
if (lex_file != (char *)NULL)
    {
    myfree(lex_file);
    lex_file = (char *)NULL;
    }
}

/*****************************************************************************
 * NAME
 *    do_recognize
 * ARGUMENTS
 *    form - pointer to a surface form
 * DESCRIPTION
 *    Execute a RECOGNIZE [<form>] command.
 *    If a form is given, generate the corresponding lexical form.  Otherwise,
 *    prompt for surface forms, generating the lexical forms, until the user
 *    types a blank line.
 * RETURN VALUE
 *    none
 */
void do_recognize(form)
unsigned char *form;
{
register unsigned char *bufp;
register RESULT *resp;
unsigned char buf[MAXLINELEN+1];

if (form != (unsigned char *)NULL)
    {
    /*
     *  form given with command -- process only that form
     */
    bufp = strtok8(form, whiteSpc);
    if (Log_fp != (FILE *)NULL)
	fprintf(Log_fp, "%s\n", bufp);

    if (Timing_flag)
	start_timing();

    resp = recognizer(bufp, &Lang,Limit_flag,Trace_flag,Log_fp);

    if (Timing_flag)
	stop_timing();

    printResults( resp, stderr );
    if (Log_fp != (FILE *)NULL)
	printResults( resp, Log_fp );
    free_result(resp);

    if (Timing_flag)
	show_timing();

    return;
    }
for (;;)
    {
    /*
     *  no form given with command -- read and process forms until blank line
     */
    fputs( "recognizer>>", stderr);
    cgetline(buf,MAXLINELEN);
    bufp = strtok8(buf, whiteSpc);
    if (bufp == (unsigned char *)NULL)
	return;			/* quit at first blank line */
    if (Log_fp != (FILE *)NULL)
	fprintf(Log_fp, "%s\n", bufp);

    if (Timing_flag)
	start_timing();

    resp = recognizer(bufp, &Lang,Limit_flag,Trace_flag,Log_fp);

    if (Timing_flag)
	stop_timing();

    printResults( resp, stderr );
    if (Log_fp != (FILE *)NULL)
	printResults( resp, Log_fp );
    free_result(resp);

    if (Timing_flag)
	show_timing();
    }
}


/****************************************************************************
 * NAME
 *    list_lexicon
 * ARGUMENTS
 *    none
 * DESCRIPTION
 *    Execute a LIST LEXICON command.
 * RETURN VALUE
 *    none
 */
void list_lexicon()
{
register int i;
    
if (Lang.num_lex_sections)
    putc('\n', stderr);
for ( i = 0 ; i < Lang.num_lex_sections ; ++i )
    fprintf(stderr, "    %s\n", Lang.lex_sections[i].lex_name);
if (Lang.num_lex_sections)
    putc('\n', stderr);
}

/*****************************************************************************
 * NAME
 *    show_lex_items
 * ARGUMENTS
 *    lip    - pointer to linked list of struct lex_item
 * DESCRIPTION
 *    Display the contents of a linked list struct lex_item's which are
 *    stored in a TRIE.
 * RETURN VALUE
 *    none
 */
static void show_lex_items(lip)
struct lex_item *lip;
{
short *xp;
unsigned char *np;
ALTERNATION *altp;

for ( ; lip != (struct lex_item *)NULL ; lip = lip->link )
    {
    fprintf(stderr, "%11s   \"%s\"   { ",
	(lip->lex_string != (unsigned char *)NULL) ?
			lip->lex_string : (unsigned char *)"0",
	(lip->lex_gloss  != (unsigned char *)NULL) ?
			lip->lex_gloss  : (unsigned char *)"" );
    if ((lip->lex_continue >= 0) && (lip->lex_continue < Lang.num_alterns))
	altp = &Lang.alterns[lip->lex_continue];
    else
	altp = (ALTERNATION *)NULL;
    if (altp != (ALTERNATION *)NULL) 
	{
	for ( xp = altp->alt_lexicons ; *xp != -1 ; ++xp )
	    {
	    np = Lang.lex_sections[*xp].lex_name;
	    fprintf(stderr, "%s ",
		    (np != (unsigned char *)NULL) ? np : (unsigned char *)"" );
	    }
	}
    else
	fprintf(stderr, "%c ", Lang.boundary );
    fprintf(stderr, "}\n" );
    }
}

/*****************************************************************************
 * NAME
 *    show_lexicon
 * ARGUMENTS
 *    lexname - name of the (sub)lexicon to show
 * DESCRIPTION
 *    Display the contents of a specified sublexicon.
 * RETURN VALUE
 *    none
 */
void show_lexicon(lexname)
unsigned char *lexname;
{
LEXICON *lxp;
    
if (lexname == (unsigned char *)NULL)
    {
    report_error(FATAL, &Miss_argument, (int *)NULL, (char *)NULL,
							"SHOW LEXICON" );
    return;
    }
lxp = find_lexicon(lexname,&Lang);
if (lxp == (LEXICON *)NULL)
    {
    report_error(FATAL, &Bad_argument, (int *)NULL, (char *)NULL,
						"SHOW LEXICON", lexname);
    return;
    }
putc('\n', stderr);
show_trie( lxp->lex_storage, show_lex_items);
putc('\n', stderr);
}

/************************************************************************
 * NAME
 *    strlower
 * ARGUMENTS
 *    string - address of string to convert to lower case
 * DESCRIPTION
 *    Convert all letters in a string to lower case.
 * RETURN VALUE
 *    original value of input argument
 */
char *strlower(string)
char *string;
{
register char *p;
register char c;

if (string != (char *)NULL)
    {
    for ( p = string ; (c = *p) != NUL ; ++p )
	{
	if (isascii(c) && isupper(c))
	    *p = tolower(c);
	}
    }
return(string);
}

#ifdef THINK_C
/*****************************************************************************
 * NAME
 *    do_cd
 * ARGUMENTS
 *    filename - name of a directory to change to
 * DESCRIPTION
 *    Make the indicated directory current.
 * RETURN VALUE
 *    none
 */
void do_cd(filename)
char *filename;
{
	short vRefNum;
	char buf[256];
	char *p;
	WDPBRec paramBlock;
	long ioWDProcID;
	
if (filename == (char *)NULL)
    {
    report_error(FATAL, &Missing_directory, (int *)NULL, (char *)NULL);
    return;
    }
if (wanthelp(filename,"CD <pathname>   (no default)\n"))
    return;

strcpy(buf, filename);
/* a directory pathname should end in a colon; supply if needed */
for(p=buf;*p;p++)
	;
if(*(p-1) != ':')
	{
	*p++ = ':';
	*p = '\0';
	}
/* The incredible mess that follows here is because the high-level
 * SetVol does not work when passed the name of a directory not
 * previously used.  Instead, it is necessary to use the low level
 * calls, and ask for a working directory to be opened. */
paramBlock.ioCompletion = (ProcPtr) NULL;
paramBlock.ioNamePtr = (StringPtr) CtoPstr(buf);
paramBlock.ioVRefNum = 0;
paramBlock.ioWDDirID = 0;
if(PBHSetVol(&paramBlock, FALSE) != 0)
	{
    report_error(FATAL, &Directory_not_found, (int *)NULL, (char *)NULL);
    return;
    }

paramBlock.ioCompletion = (ProcPtr) NULL;
paramBlock.ioNamePtr = NULL;
PBHGetVol (&paramBlock, FALSE);
ioWDProcID = paramBlock.ioWDProcID;
paramBlock.ioNamePtr = NULL;
PBOpenWD (&paramBlock, FALSE);
SetVol(NULL, paramBlock.ioVRefNum);
}
#endif  /*THINK_C*/

#ifndef THINK_C
/*****************************************************************************
 * NAME
 *    do_edit
 * ARGUMENTS
 *    filename - name of a file to edit
 * DESCRIPTION
 *    Try to edit the indicated file with the user's preferred editor.
 * RETURN VALUE
 *    none
 */
void do_edit(filename)
char *filename;
{
char cmd[200];
char *editor;

if (filename == (char *)NULL)
    {
    report_error(FATAL, &Missing_edit_file, (int *)NULL, (char *)NULL);
    return;
    }
if (wanthelp(filename,"EDIT <file> (no default filename or type)"))
    return;

editor = getenv("EDITOR");
if (editor == (char *)NULL)
    {
#ifdef MSDOS
    editor = "EDLIN";
#endif
#ifdef UNIX
    editor = "vi";
#endif
    }
sprintf(cmd, "%s %s", editor, filename);
if (system(cmd) < 0)
    perror("PCKIMMO");
}
#endif  /*THINK_C*/

/*****************************************************************************
 * NAME
 *    do_
 * ARGUMENTS
 *    
 * DESCRIPTION
 *    
 * RETURN VALUE
 *    none
 */
